import ujson as json
from mock import patch
from tornado.testing import AsyncHTTPTestCase

from tests.conftest import MockBaseHandler


class TestGenerateChangesHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_post_no_user(self):
        expected = b"No user detected. Check configuration."
        response = self.fetch("/api/v2/generate_changes", method="POST", body="abcd")
        self.assertEqual(response.code, 403)
        self.assertEqual(response.body, expected)

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_invalid_requests(self):
        input_body = {"changes": [{"arn": "arn:aws:s3::123456789012:example_bucket"}]}
        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertIn("Error validating input", str(response.body))
        self.assertEqual(response.code, 400)

        input_body["changes"][0][
            "resource"
        ] = "arn:aws:s3::12345678902:example_bucket_2"
        input_body["changes"][0]["generator_type"] = "fake"

        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertIn("Error validating input", str(response.body))
        self.assertEqual(response.code, 400)

        input_body["changes"][0]["generator_type"] = "s3"
        input_body["changes"][0]["action_groups"] = ["get", "fakeaction"]

        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertIn("Error validating input", str(response.body))
        self.assertEqual(response.code, 400)

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_generic(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "s3",
                    "resource_arn": "arn:aws:s3:::123456789012-bucket",
                    "bucket_prefix": "/*",
                    "effect": "Allow",
                    "action_groups": ["get", "list"],
                },
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "s3",
                    "resource_arn": "arn:aws:s3:::bucket2",
                    "bucket_prefix": "/*",
                    "effect": "Allow",
                    "action_groups": ["list", "get"],
                },
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "crud_lookup",
                    "resource_arn": "*",
                    "effect": "Allow",
                    "service_name": "ssm",
                    "action_groups": ["list", "read"],
                },
            ]
        }

        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            input_body["changes"][0]["principal"]["principal_arn"],
        )
        self.assertEqual(
            len(result["changes"]),
            1,
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_wildcard(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "s3",
                    "resource_arn": "*",
                    "bucket_prefix": "folder_name/filename",
                    "effect": "Allow",
                    "action_groups": ["get", "list"],
                },
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "s3",
                    "resource_arn": "*",
                    "bucket_prefix": "folder_name/*",
                    "effect": "Allow",
                    "action_groups": ["list", "get"],
                },
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "crud_lookup",
                    "resource_arn": "*",
                    "effect": "Allow",
                    "service_name": "ssm",
                    "action_groups": ["list", "read"],
                },
            ]
        }

        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            input_body["changes"][0]["principal"]["principal_arn"],
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_s3(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "resource_arn": "arn:aws:s3::123456789012:examplebucket",
                    "bucket_prefix": "/*",
                    "generator_type": "s3",
                    "version": "abcd",
                    "asd": "sdf",
                    "action_groups": ["list", "delete"],
                }
            ]
        }
        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            "arn:aws:iam::123456789012:role/roleName",
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_s3_multi(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "resource_arn": [
                        "arn:aws:s3::123456789012:examplebucket",
                        "arn:aws:s3::123456789012:examplebucket2",
                    ],
                    "bucket_prefix": "/*",
                    "generator_type": "s3",
                    "version": "abcd",
                    "asd": "sdf",
                    "action_groups": ["list", "delete"],
                }
            ]
        }
        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            "arn:aws:iam::123456789012:role/roleName",
        )
        del result["changes"][0]["policy"]["policy_document"]["Statement"][0]["Sid"]
        self.assertEqual(
            result["changes"][0]["policy"]["policy_document"]["Statement"],
            [
                {
                    "Action": [
                        "s3:deleteobject",
                        "s3:deleteobjecttagging",
                        "s3:deleteobjectversion",
                        "s3:deleteobjectversiontagging",
                        "s3:listbucket",
                        "s3:listbucketversions",
                    ],
                    "Effect": "Allow",
                    "Resource": [
                        "arn:aws:s3:::arn:aws:s3::123456789012:examplebucket",
                        "arn:aws:s3:::arn:aws:s3::123456789012:examplebucket/*",
                        "arn:aws:s3:::arn:aws:s3::123456789012:examplebucket2",
                        "arn:aws:s3:::arn:aws:s3::123456789012:examplebucket2/*",
                    ],
                }
            ],
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_s3_combined_inline(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "resource_arn": "arn:aws:s3::123456789012:examplebucket",
                    "bucket_prefix": "/*",
                    "generator_type": "s3",
                    "version": "abcd",
                    "asd": "sdf",
                    "action_groups": ["list", "delete"],
                },
                {
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "custom_iam",
                    "policy": {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "IncludeAccounts": [
                                    "account_a",
                                    "account_b",
                                    "account_c",
                                ],
                                "Action": [
                                    "s3:GetObjectVersion",
                                    "s3:GetObject",
                                    "s3:GetObjectTagging",
                                    "s3:GetObjectAcl",
                                    "s3:ListBucket",
                                    "s3:GetObjectVersionAcl",
                                    "s3:ListBucketVersions",
                                    "s3:GetObjectVersionTagging",
                                ],
                                "Effect": "Allow",
                                "Resource": [
                                    "arn:aws:s3:::bucket2",
                                    "arn:aws:s3:::bucket2/*",
                                ],
                            }
                        ],
                    },
                },
            ]
        }
        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            "arn:aws:iam::123456789012:role/roleName",
        )
        policy = result["changes"][0]["policy"]["policy_document"]
        policy["Statement"][0].pop("Sid")
        policy["Statement"][1].pop("Sid")
        self.assertEqual(
            policy["Statement"][0],
            {
                "Action": [
                    "s3:deleteobject",
                    "s3:deleteobjecttagging",
                    "s3:deleteobjectversion",
                    "s3:deleteobjectversiontagging",
                    "s3:listbucket",
                    "s3:listbucketversions",
                ],
                "Effect": "Allow",
                "Resource": [
                    "arn:aws:s3:::arn:aws:s3::123456789012:examplebucket",
                    "arn:aws:s3:::arn:aws:s3::123456789012:examplebucket/*",
                ],
            },
        )

        self.assertEqual(
            policy["Statement"][1],
            {
                "Action": [
                    "s3:getobject",
                    "s3:getobjectacl",
                    "s3:getobjecttagging",
                    "s3:getobjectversion",
                    "s3:getobjectversionacl",
                    "s3:getobjectversiontagging",
                    "s3:listbucket",
                    "s3:listbucketversions",
                ],
                "Effect": "Allow",
                "IncludeAccounts": ["account_a", "account_b", "account_c"],
                "Resource": ["arn:aws:s3:::bucket2", "arn:aws:s3:::bucket2/*"],
            },
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    @patch("consoleme.handlers.v2.generate_changes.ChangeGeneratorModelArray.parse_raw")
    def test_post_raises(self, mock_change_generator_model_array_parse_raw):
        mock_change_generator_model_array_parse_raw.side_effect = Exception(
            "Unknown Exception!"
        )
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "resource_arn": "arn:aws:s3::123456789012:examplebucket",
                    "bucket_prefix": "/*",
                    "generator_type": "s3",
                    "version": "abcd",
                    "asd": "sdf",
                    "action_groups": ["list", "delete"],
                }
            ]
        }
        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 500)
        self.assertIn("Error generating changes", str(response.body))

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_sns(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/exampleRole",
                    },
                    "resource_arn": "arn:aws:sns:us-east-1:123456789012:exampletopic",
                    "generator_type": "sns",
                    "version": "abcd",
                    "asd": "sdf",
                    "action_groups": ["get_topic_attributes", "publish"],
                }
            ]
        }

        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            "arn:aws:iam::123456789012:role/exampleRole",
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_sns_multi(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/exampleRole",
                    },
                    "resource_arn": [
                        "arn:aws:sns:us-east-1:123456789012:exampletopic",
                        "arn:aws:sns:us-east-1:123456789012:exampletopic2",
                        "arn:aws:sns:us-east-1:123456789012:exampletopic3",
                    ],
                    "generator_type": "sns",
                    "version": "abcd",
                    "asd": "sdf",
                    "action_groups": ["get_topic_attributes", "publish"],
                }
            ]
        }

        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            "arn:aws:iam::123456789012:role/exampleRole",
        )

        del result["changes"][0]["policy"]["policy_document"]["Statement"][0]["Sid"]
        self.assertEqual(
            result["changes"][0]["policy"]["policy_document"]["Statement"],
            [
                {
                    "Action": [
                        "sns:getendpointattributes",
                        "sns:gettopicattributes",
                        "sns:publish",
                    ],
                    "Effect": "Allow",
                    "Resource": [
                        "arn:aws:sns:us-east-1:123456789012:exampletopic",
                        "arn:aws:sns:us-east-1:123456789012:exampletopic2",
                        "arn:aws:sns:us-east-1:123456789012:exampletopic3",
                    ],
                }
            ],
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_post_valid_request_sqs(self):
        input_body = {
            "changes": [
                {
                    "user": "username@example.com",
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/roleName",
                    },
                    "generator_type": "sqs",
                    "resource_arn": "arn:aws:sqs:us-east-1:123456789012:resourceName",
                    "effect": "Allow",
                    "action_groups": ["get_queue_attributes", "send_messages"],
                }
            ]
        }
        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["principal_arn"],
            "arn:aws:iam::123456789012:role/roleName",
        )

    @patch(
        "consoleme.handlers.v2.generate_changes.GenerateChangesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_generate_changes_honeybee_template(self):
        input_body = {
            "changes": [
                {
                    "principal": {
                        "principal_type": "HoneybeeAwsResourceTemplate",
                        "repository_name": "honeybee_templates",
                        "resource_identifier": "iamrole/abc.yaml",
                        "resource_url": "http://example.com/fake_repo/path/to/file.yaml",
                    },
                    "generator_type": "s3",
                    "action_groups": ["list", "get"],
                    "extra_actions": ["s3:get*"],
                    "effect": "Allow",
                    "resource_arn": "arn:aws:s3:::bucketa",
                    "bucket_prefix": "/*",
                    "include_accounts": ["account_a", "account_b"],
                },
                {
                    "principal": {
                        "principal_type": "HoneybeeAwsResourceTemplate",
                        "repository_name": "honeybee_templates",
                        "resource_identifier": "iamrole/abc.yaml",
                        "resource_url": "http://example.com/fake_repo/path/to/file.yaml",
                    },
                    "generator_type": "custom_iam",
                    "policy": {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "IncludeAccounts": ["account_a", "account_b"],
                                "Action": [
                                    "s3:GetObjectVersion",
                                    "s3:GetObject",
                                    "s3:GetObjectTagging",
                                    "s3:GetObjectAcl",
                                    "s3:ListBucket",
                                    "s3:GetObjectVersionAcl",
                                    "s3:ListBucketVersions",
                                    "s3:GetObjectVersionTagging",
                                ],
                                "Effect": "Allow",
                                "Resource": [
                                    "arn:aws:s3:::bucketb",
                                    "arn:aws:s3:::bucketb/prefix/*",
                                ],
                            }
                        ],
                    },
                },
            ]
        }
        response = self.fetch(
            "/api/v2/generate_changes", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(
            result["changes"][0]["principal"]["resource_identifier"],
            "iamrole/abc.yaml",
        )
        self.assertEqual(
            result["changes"][0]["resources"],
            [
                {
                    "arn": "arn:aws:s3:::bucketa",
                    "name": "bucketa",
                    "account_id": "",
                    "region": "global",
                    "account_name": "",
                    "policy_sha256": None,
                    "policy": None,
                    "actions": None,
                    "owner": None,
                    "approvers": None,
                    "resource_type": "s3",
                    "last_updated": None,
                }
            ],
        )

        result["changes"][0]["policy"]["policy_document"]["Statement"][0].pop("Sid")
        result["changes"][0]["policy"]["policy_document"]["Statement"][1].pop("Sid")
        self.assertEqual(
            result["changes"][0]["policy"]["policy_document"]["Statement"],
            [
                {
                    "Action": ["s3:get*", "s3:listbucket", "s3:listbucketversions"],
                    "Effect": "Allow",
                    "IncludeAccounts": ["account_a", "account_b"],
                    "Resource": ["arn:aws:s3:::bucketa", "arn:aws:s3:::bucketa/*"],
                },
                {
                    "Action": [
                        "s3:getobject",
                        "s3:getobjectacl",
                        "s3:getobjecttagging",
                        "s3:getobjectversion",
                        "s3:getobjectversionacl",
                        "s3:getobjectversiontagging",
                        "s3:listbucket",
                        "s3:listbucketversions",
                    ],
                    "Effect": "Allow",
                    "IncludeAccounts": ["account_a", "account_b"],
                    "Resource": [
                        "arn:aws:s3:::bucketb",
                        "arn:aws:s3:::bucketb/prefix/*",
                    ],
                },
            ],
        )



import sys
from datetime import datetime, timedelta

import sentry_sdk
import ujson as json
from policy_sentry.util.arns import parse_arn

from consoleme.config import config
from consoleme.exceptions.exceptions import MustBeFte, ResourceNotFound
from consoleme.handlers.base import BaseAPIV2Handler, BaseMtlsHandler
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.auth import can_admin_policies
from consoleme.lib.aws import fetch_resource_details
from consoleme.lib.cache import retrieve_json_data_from_redis_or_s3
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.policies import get_url_for_resource
from consoleme.lib.redis import RedisHandler, redis_hget
from consoleme.lib.web import handle_generic_error_response
from consoleme.models import WebResponse

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
group_mapping = get_plugin_by_name(
    config.get("plugins.group_mapping", "default_group_mapping")
)()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
internal_policies = get_plugin_by_name(
    config.get("plugins.internal_policies", "default_policies")
)()
red = RedisHandler().redis_sync()


class ResourceDetailHandler(BaseAPIV2Handler):
    async def get(self, account_id, resource_type, region=None, resource_name=None):
        if not self.user:
            return
        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")
        read_only = False

        can_save_delete = (can_admin_policies(self.user, self.groups),)

        account_id_for_arn: str = account_id
        if resource_type == "s3":
            account_id_for_arn = ""
        arn = f"arn:aws:{resource_type}:{region or ''}:{account_id_for_arn}:{resource_name}"
        path = ""
        if resource_type == "managed_policy":
            # special case for managed policies
            path = region or ""
            if path:
                arn = f"arn:aws:iam::{account_id}:policy/{path}/{resource_name}"
            else:
                arn = f"arn:aws:iam::{account_id}:policy/{resource_name}"

        stats.count(
            "ResourcePolicyEditHandler.get", tags={"user": self.user, "arn": arn}
        )

        log_data = {
            "user": self.user,
            "ip": self.ip,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "message": "Incoming request",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "arn": arn,
        }

        log.debug(log_data)

        error = ""

        try:
            resource_details = await fetch_resource_details(
                account_id, resource_type, resource_name, region, path
            )
        except Exception as e:
            sentry_sdk.capture_exception()
            log.error({**log_data, "error": e}, exc_info=True)
            resource_details = None
            error = str(e)

        if not resource_details:
            self.send_error(
                404,
                message=(
                    f"Unable to retrieve the specified {resource_type} resource: "
                    f"{account_id}/{resource_name}/{region}. {error}",
                ),
            )
            return

        # TODO: Get S3 errors for s3 buckets only, else CT errors
        yesterday = (datetime.today() - timedelta(days=1)).strftime("%Y%m%d")
        s3_query_url = None
        if resource_type == "s3":
            s3_query_url = config.get("s3.bucket_query_url")
        all_s3_errors = None
        if s3_query_url:
            s3_query_url = s3_query_url.format(
                yesterday=yesterday, bucket_name=f"'{resource_name}'"
            )
            s3_error_topic = config.get("redis.s3_errors", "S3_ERRORS")
            all_s3_errors = self.red.get(s3_error_topic)

        s3_errors = []
        if all_s3_errors:
            s3_errors = json.loads(all_s3_errors).get(arn, [])

        account_ids_to_name = await get_account_id_to_name_mapping()
        # TODO(ccastrapel/psanders): Make a Swagger spec for this
        self.write(
            dict(
                arn=arn,
                resource_details=resource_details,
                account_id=account_id,
                account_name=account_ids_to_name.get(account_id, None),
                read_only=read_only,
                can_save_delete=can_save_delete,
                s3_errors=s3_errors,
                error_url=s3_query_url,
                config_timeline_url=resource_details.get("config_timeline_url"),
            )
        )


class GetResourceURLHandler(BaseMtlsHandler):
    """consoleme CLI resource URL handler. Parameters accepted: arn."""

    def initialize(self):
        self.user: str = None
        self.eligible_roles: list = []

    async def get(self):
        """
        /api/v2/get_resource_url - Endpoint used to get an URL from an ARN
        ---
        get:
            description: Get the resource URL for ConsoleMe, given an ARN
            responses:
                200:
                    description: Returns a URL generated from the ARN in JSON form
                400:
                    description: Malformed Request
                403:
                    description: Forbidden
        """
        self.user: str = self.requester["email"]
        arn: str = self.get_argument("arn", None)
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": self.user,
            "arn": arn,
            "message": "Generating URL for resource",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }
        log.debug(log_data)
        stats.count("GetResourceURL.get", tags={"user": self.user})
        if not arn:
            generic_error_message: str = "Missing required parameter"
            errors = ["arn is a required parameter"]
            await handle_generic_error_response(
                self, generic_error_message, errors, 404, "missing_data", log_data
            )
            return

        try:
            # parse_arn will raise an exception on invalid arns
            parse_arn(arn)

            resources_from_aws_config_redis_key = config.get(
                "aws_config_cache.redis_key", "AWSCONFIG_RESOURCE_CACHE"
            )
            if not red.exists(resources_from_aws_config_redis_key):
                # This will force a refresh of our redis cache if the data exists in S3
                await retrieve_json_data_from_redis_or_s3(
                    redis_key=resources_from_aws_config_redis_key,
                    s3_bucket=config.get("aws_config_cache_combined.s3.bucket"),
                    s3_key=config.get(
                        "aws_config_cache_combined.s3.file",
                        "aws_config_cache_combined/aws_config_resource_cache_combined_v1.json.gz",
                    ),
                    redis_data_type="hash",
                )
            resource_info = await redis_hget(resources_from_aws_config_redis_key, arn)
            if not resource_info:
                raise ValueError("Resource not found in organization cache")
            resource_url = await get_url_for_resource(arn)
            if not resource_url:
                raise ValueError("This resource type is currently not supported")
        except (ResourceNotFound, ValueError) as e:
            generic_error_message: str = "Unsupported data"
            errors = [str(e)]
            await handle_generic_error_response(
                self, generic_error_message, errors, 404, "invalid_data", log_data
            )
            return
        except Exception as e:
            generic_error_message: str = "Malformed data"
            errors = [str(e)]
            await handle_generic_error_response(
                self, generic_error_message, errors, 404, "malformed_data", log_data
            )
            return

        res = WebResponse(
            status="success",
            status_code=200,
            message="Successfully generated URL for ARN",
            data={"url": resource_url},
        )

        self.write(res.json())
        await self.finish()



import sys
from typing import Dict

from consoleme.config import config
from consoleme.lib.cloud_credential_authorization_mapping.models import (
    CredentialAuthzMappingGenerator,
    RoleAuthorizations,
    user_or_group,
)
from consoleme.lib.redis import RedisHandler

red = RedisHandler().redis_sync()


class DynamicConfigAuthorizationMappingGenerator(CredentialAuthzMappingGenerator):
    async def generate_credential_authorization_mapping(
        self, authorization_mapping: Dict[user_or_group, RoleAuthorizations]
    ) -> Dict[user_or_group, RoleAuthorizations]:
        """This will list accounts that meet the account attribute search criteria."""
        function = f"{__name__}.{sys._getframe().f_code.co_name}"
        log_data = {
            "function": function,
        }
        config.CONFIG.load_dynamic_config_from_redis(log_data, red)
        group_mapping_configuration = config.get("dynamic_config.group_mapping")

        if not group_mapping_configuration:
            return authorization_mapping

        for group, role_mapping in group_mapping_configuration.items():
            if config.get("auth.force_groups_lowercase", False):
                group = group.lower()
            if not authorization_mapping.get(group):
                authorization_mapping[group] = RoleAuthorizations.parse_obj(
                    {"authorized_roles": set(), "authorized_roles_cli_only": set()}
                )
            authorization_mapping[group].authorized_roles.update(
                role_mapping.get("roles", [])
            )
            authorization_mapping[group].authorized_roles_cli_only.update(
                role_mapping.get("cli_only_roles", [])
            )
        return authorization_mapping



"""
Spoke accounts stack for running ConsoleMe on ECS
"""

from aws_cdk import aws_iam as iam
from aws_cdk import core as cdk

from consoleme_ecs_cdk.service.constants import MAIN_ACCOUNT_ID, SPOKE_BASE_NAME


class ConsolemeSpokeAccountsStack(cdk.Stack):
    """
    Spoke accounts stack for running ConsoleMe on ECS
    Granting the necessary permissions for ConsoleMe main account role
    """

    def __init__(self, scope: cdk.Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        trusted_role_arn = "arn:aws:iam::" + MAIN_ACCOUNT_ID + ":role/ConsoleMeTaskRole"

        spoke_role = iam.Role(
            self,
            f"{SPOKE_BASE_NAME}TrustRole",
            role_name="ConsoleMeTrustRole",
            assumed_by=iam.ArnPrincipal(arn=trusted_role_arn),
        )

        spoke_role.add_to_policy(
            iam.PolicyStatement(
                effect=iam.Effect.ALLOW,
                actions=[
                    "autoscaling:Describe*",
                    "cloudwatch:Get*",
                    "cloudwatch:List*",
                    "config:BatchGet*",
                    "config:List*",
                    "config:Select*",
                    "ec2:describeregions",
                    "ec2:DescribeSubnets",
                    "ec2:describevpcendpoints",
                    "ec2:DescribeVpcs",
                    "iam:*",
                    "s3:GetBucketPolicy",
                    "s3:GetBucketTagging",
                    "s3:ListAllMyBuckets",
                    "s3:ListBucket",
                    "s3:PutBucketPolicy",
                    "s3:PutBucketTagging",
                    "sns:GetTopicAttributes",
                    "sns:ListTagsForResource",
                    "sns:ListTopics",
                    "sns:SetTopicAttributes",
                    "sns:TagResource",
                    "sns:UnTagResource",
                    "sqs:GetQueueAttributes",
                    "sqs:GetQueueUrl",
                    "sqs:ListQueues",
                    "sqs:ListQueueTags",
                    "sqs:SetQueueAttributes",
                    "sqs:TagQueue",
                    "sqs:UntagQueue",
                ],
                resources=["*"],
            )
        )



import asyncio
import html
import sys
from typing import Any, Dict, List, Optional, Union

import googleapiclient.discovery
import ujson as json
from asgiref.sync import sync_to_async
from google.oauth2 import service_account
from googleapiclient.discovery import Resource
from googleapiclient.errors import HttpError
from retrying import retry
from validate_email import validate_email

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    BackgroundCheckNotPassedException,
    BulkAddPrevented,
    DifferentUserGroupDomainException,
    MissingConfigurationValue,
    NoCredentialSubjectException,
    NoGroupsException,
    NotAMemberException,
    UnableToModifyRestrictedGroupMembers,
    UnauthorizedToAccess,
    UserAlreadyAMemberOfGroupException,
)
from consoleme.lib.auth import can_modify_members
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.groups import does_group_require_bg_check
from consoleme.lib.plugins import get_plugin_by_name

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()

log = config.get_logger()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()


async def add_user_to_group_task(
    member: str,
    group: str,
    requesting_user: str,
    requesting_users_groups: List[str],
    semaphore=None,
    service: None = None,
) -> Dict[str, Union[str, bool]]:
    if not semaphore:
        semaphore = asyncio.BoundedSemaphore(10)
    async with semaphore:
        stats.count(
            "add_user_to_group_task.attempt",
            tags={"member": member, "group": group, "requesting_user": requesting_user},
        )
        member = member.strip()
        result = {
            "Action": "Add user",
            "Member": member,
            "Group": group,
            "Error": False,
        }
        log_data = {
            "function": f"{__name__, sys._getframe().f_code.co_name}",
            "action": "Add user",
            "member": member,
            "group": group,
        }
        try:
            group_info = await auth.get_group_info(group, members=False)
            can_add_remove_members = can_modify_members(
                requesting_user, requesting_users_groups, group_info
            )

            if not can_add_remove_members:
                result[
                    "Result"
                ] = "You are unable to add members to this group. Maybe it is restricted."
                result["Error"] = True
                error = f"There was at least one problem. {result['Result']}"
                log_data["error"] = error
                log.warning(log_data, exc_info=True)

                return result
            if not validate_email(member):
                result["Result"] = "Invalid e-mail address entered"
                result["Error"] = True
                log_data["message"] = "Error"
                log_data["error"] = result["Result"]
                log.warning(log_data, exc_info=True)
                return result

            if (
                not group_info.allow_third_party_users
                and not await auth.does_user_exist(member)
            ):
                result[
                    "Result"
                ] = "User does not exist in our environment and this group doesn't allow third party users."
                result["Error"] = True
                log_data["message"] = "Error"
                log_data["error"] = result["Result"]
                log.warning(log_data, exc_info=True)
                return result

            await add_user_to_group(member, group, requesting_user, service=service)
            result["Result"] = "Successfully added user to group"
            return result
        except Exception as e:
            result["Result"] = html.escape(str(e))
            result["Error"] = True
            error = f"There was at least one problem. {e}"
            log_data["message"] = "Error"
            log_data["error"] = error
            log.error(log_data, exc_info=True)
            return result


async def remove_user_from_group_task(
    member: str,
    group: str,
    requesting_user: str,
    requesting_users_groups: List[str],
    semaphore=None,
    service: None = None,
) -> Dict[str, Union[str, bool]]:
    if not semaphore:
        semaphore = asyncio.BoundedSemaphore(10)
    async with semaphore:
        stats.count(
            "remove_user_from_group_task.attempt",
            tags={"member": member, "group": group, "requesting_user": requesting_user},
        )
        member = member.strip()
        result = {
            "Action": "Remove user",
            "Member": member,
            "Requesting User": requesting_user,
            "Group": group,
            "Error": False,
        }
        log_data = {
            "function": f"{__name__, sys._getframe().f_code.co_name}",
            "action": "Remove user",
            "member": member,
            "group": group,
        }

        try:
            group_info = await auth.get_group_info(group, members=False)
            can_add_remove_members = can_modify_members(
                requesting_user, requesting_users_groups, group_info
            )

            if not can_add_remove_members:
                result[
                    "Result"
                ] = "You are unable to remove members from this group. Maybe it is restricted."
                result["Error"] = True
                error = f"There was at least one problem. {result['Result']}"
                log_data["error"] = error
                log.warning(log_data, exc_info=True)

                return result

            if not validate_email(member):
                result[
                    "Result"
                ] = "Invalid e-mail address entered, or user doesn't exist"
                result["Error"] = True
                log_data["message"] = "Error"
                log_data["error"] = result["Result"]
                log.warning(log_data, exc_info=True)
                return result

            await remove_user_from_group(
                member, group, requesting_user, service=service
            )
            result["Result"] = "Successfully removed user from group"
            return result
        except Exception as e:
            result["Result"] = str(e)
            result["Error"] = True
            error = f"There was at least one problem. {e}"
            log_data["message"] = "Error"
            log_data["error"] = error
            log.error(log_data, exc_info=True)
            return result


async def get_service(service_name: str, service_path: str, group: str) -> Resource:
    """
    Get a service connection to Google. You'll need to generate a GCP service account first from instructions here:
    https://hawkins.gitbook.io/consoleme/configuration/authentication-and-authorization/google-groups-support

    ConsoleMe requires that you either have a service key file with content like below,
    and you've set the configuration for `google.service_key_file` to the full path of that file on disk,
    or you've just put the json for this in your ConsoleMe configuration in the `google.service_key_dict` configuration
    key.

    There are sensitive secrets here, so if you want to
    reference them directly in configuration, we encourage you to store these secrets in AWS Secrets Manager
    https://hawkins.gitbook.io/consoleme/configuration/aws-secret-manager-integration
        {
          "type": "service_account",
          "project_id": "cons...",
          "private_key_id": "dc61.....",
          "private_key": "-----BEGIN PRIVATE KEY-----\nMII.....",
          "client_email": "cons...@cons....gserviceaccount.com",
          "client_id": "1234....",
          "auth_uri": "https://accounts.google.com/o/oauth2/auth",
          "token_uri": "https://oauth2.googleapis.com/token",
          "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
          "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/consolem...."
        }

    """

    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "service_name": service_name,
        "service_path": service_path,
        "group": group,
        "message": f"Building service connection for {service_name} / {service_path}",
    }
    log.debug(log_data)
    if config.get("google.service_key_file"):
        admin_credentials = service_account.Credentials.from_service_account_file(
            config.get("google.service_key_file"),
            scopes=config.get(
                "google.admin_scopes",
                ["https://www.googleapis.com/auth/admin.directory.group"],
            ),
        )
    elif config.get("google.service_key_dict"):
        admin_credentials = service_account.Credentials.from_service_account_info(
            config.get("google.service_key_dict"),
            scopes=config.get(
                "google.admin_scopes",
                ["https://www.googleapis.com/auth/admin.directory.group"],
            ),
        )
    else:
        raise MissingConfigurationValue(
            "Missing configuration for Google. You must configure either `google.service_key_file` "
            "or `google.service_key_dict`."
        )

    # Change credential subject based on group domain
    credential_subjects = config.get("google.credential_subject")
    credential_subject = None
    for k, v in credential_subjects.items():
        if k == group.split("@")[1]:
            credential_subject = v
            break

    if not credential_subject:
        raise NoCredentialSubjectException(
            "Error: Unable to find Google credential subject for domain {}. "
            "{}".format(group.split("@")[1], config.get("ses.support_reference", ""))
        )

    admin_delegated_credentials = admin_credentials.with_subject(credential_subject)
    service = await sync_to_async(googleapiclient.discovery.build)(
        service_name, service_path, credentials=admin_delegated_credentials
    )

    return service


@sync_to_async
def list_group_members_call(service, email):
    return service.members().list(groupKey=email).execute()


@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000)
async def list_group_members(
    email: str, dry_run: None = None, service: Optional[Resource] = None
) -> List[str]:
    """List all members of a group."""
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "user": email,
        "message": "Getting list of members for group",
    }
    log.debug(log_data)
    if not service:
        service = await get_service("admin", "directory_v1", email)

    if not dry_run:
        try:
            results = await list_group_members_call(service, email)
        except HttpError as he:
            errors = json.loads(he.content.decode())
            log.debug(errors)
            raise he
        return list(map(lambda x: x.get("email", ""), results.get("members", [])))

    return []


@sync_to_async
def list_user_groups_call(service, user_email, page_token=None):
    if page_token:
        results = (
            service.groups().list(userKey=user_email, pageToken=page_token).execute()
        )
    else:
        results = service.groups().list(userKey=user_email).execute()
    return results


async def get_group_memberships(user_email, dry_run=None, service=None):
    """Get group memberships for a user"""
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "user": user_email,
        "message": "Getting list of groups for user",
    }
    log.debug(log_data)
    if not service:
        service = await get_service("admin", "directory_v1", user_email)
    groups = []
    if not dry_run:
        try:
            page_token = None
            while True:
                results = await list_user_groups_call(service, user_email, page_token)
                for g in results.get("groups"):
                    groups.append(g.get("email"))
                page_token = results.get("nextPageToken")
                if not page_token:
                    break
        except HttpError as he:
            errors = json.loads(he.content.decode())
            log.debug(errors)
            raise he
        return groups

    return []


async def raise_if_requires_bgcheck_and_no_bgcheck(user: str, group_info: Any) -> bool:
    """Check if group requires a background check, and if the user has completed the
    background check. Will raise if the user requires a background check but has not
    completed one.
    """
    if not does_group_require_bg_check(group_info):
        return True

    user_info = await auth.get_user_info(user, object=True)

    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "user": user,
        "group": group_info.name,
        "backgroundcheck_required": group_info.backgroundcheck_required,
    }
    log.debug(log_data)

    if user_info.passed_background_check:
        return True
    raise BackgroundCheckNotPassedException(
        config.get(
            "google.background_check_fail_message",
            "User {user} has not passed background check Group {group_name} requires a background check.",
        ).format(user=user, group_name=group_info.name)
    )


async def raise_if_not_same_domain(user: str, group_info: Any) -> None:
    """Check if user is in the same domain as the group. Consoleme will refuse to add users to groups under a different
    domain if the allow_cross_domain_users or allow_third_party_users attributes are not set to "true" for a group.
    """
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "user": user,
        "group": group_info.name,
        "allow_cross_domain_users": group_info.allow_cross_domain_users,
    }
    log.debug(log_data)

    if group_info.allow_cross_domain_users:
        return True

    if group_info.allow_third_party_users:
        return True

    if user.split("@")[1] != group_info.name.split("@")[1]:
        raise DifferentUserGroupDomainException(
            config.get(
                "google.different_domain_fail_message",
                "Unable to add user to a group that is in a different domain. User: {user}. Group: {group_name}",
            ).format(user=user, group_name=group_info.name)
        )


async def raise_if_restricted(user: str, group_info: Any) -> None:
    """Check if the group is a restricted group. Currently, Consoleme is not able to add users to
    restricted groups.
    """
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "user": user,
        "group": group_info.name,
        "restricted": group_info.restricted,
        "compliance_restricted": group_info.compliance_restricted,
    }
    log.debug(log_data)
    if group_info.restricted:
        raise UnableToModifyRestrictedGroupMembers(
            f"Group {group_info.name} is marked as Restricted. These groups have been determined to be sensitive. "
            f"Consoleme cannot "
            f"currently be used to add/remove users from "
            f"restricted groups."
        )


async def raise_if_bulk_add_disabled_and_no_request(
    group_info: Any, request: Optional[Dict[str, Union[int, str]]]
) -> bool:
    """Check if the group has prevent_bulk_add flag. If so, a request must have been passed in for the user to be
    added to the group.
    """
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {"function": function, "group": group_info.name, "request": str(request)}
    log.debug(log_data)
    error = config.get(
        "google.bulk_add_disabled_fail_message",
        "Group {group_name} has an attribute to prevent manually adding users to it. "
        "Users must manually request access to it".format(group_name=group_info.name),
    )
    if not group_info.prevent_bulk_add:
        return True
    if not request:
        raise BulkAddPrevented(error)
    if not request["status"] == "approved":
        raise BulkAddPrevented(error)


@sync_to_async
def insert_group_members_call(service, google_group_email, user_email, role):
    return (
        service.members()
        .insert(groupKey=google_group_email, body=dict(email=user_email, role=role))
        .execute()
    )


async def add_user_to_group(
    user_email: str,
    google_group_email: str,
    updated_by: Optional[str] = None,
    role: str = "MEMBER",
    dry_run: None = None,
    service: Optional[Resource] = None,
    request: Optional[Dict[str, Union[int, str]]] = None,
) -> Dict[str, bool]:
    """Add user member to group."""
    dynamo = UserDynamoHandler()
    result = {"done": True}
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "user_email": user_email,
        "google_group_email": google_group_email,
        "updated_by": updated_by,
        "role": role,
        "dry_run": dry_run,
        "message": "Adding user to group",
    }
    if not service:
        service = await get_service("admin", "directory_v1", google_group_email)
    existing = await list_group_members(google_group_email, dry_run=dry_run)

    if user_email in existing:
        log_data["message"] = "Unable to add user to group. User is already a member."
        log.warning(log_data)
        result["done"] = False
        result["message"] = log_data["message"]
        raise UserAlreadyAMemberOfGroupException(result["message"])

    group_info = await auth.get_group_info(google_group_email, members=False)
    await raise_if_requires_bgcheck_and_no_bgcheck(user_email, group_info)
    await raise_if_not_same_domain(user_email, group_info)
    await raise_if_restricted(user_email, group_info)
    await raise_if_bulk_add_disabled_and_no_request(group_info, request)

    if not dry_run:
        stats.count(
            "google.add_user_to_group",
            tags={
                "user_email": user_email,
                "google_group_email": google_group_email,
                "updated_by": updated_by,
            },
        )

        await insert_group_members_call(service, google_group_email, user_email, role)
        await dynamo.create_group_log_entry(
            google_group_email, user_email, updated_by, "Added"
        )
        log.info(log_data)
    return result


async def api_add_user_to_group_or_raise(group_name, member_name, actor):
    try:
        group_info = await auth.get_group_info(group_name, members=False)
    except Exception:
        raise NoGroupsException("Unable to retrieve the specified group")

    actor_groups = await auth.get_groups(actor)
    can_add_remove_members = can_modify_members(actor, actor_groups, group_info)

    if not can_add_remove_members:
        raise UnauthorizedToAccess("Unauthorized to modify members of this group.")

    try:
        await add_user_to_group(member_name, group_name, actor)
    except HttpError as e:
        # Inconsistent GG API error - ignore failure for user already existing
        if e.resp.reason == "duplicate":
            pass
    except UserAlreadyAMemberOfGroupException:
        pass
    except BulkAddPrevented:
        dynamo_handler = UserDynamoHandler(actor)
        dynamo_handler.add_request(
            member_name,
            group_name,
            f"{actor} requesting on behalf of {member_name} from a bulk operation",
            updated_by=actor,
        )
        return "REQUESTED"

    return "ADDED"


@sync_to_async
def delete_group_members_call(service, google_group_email, user_email):
    return (
        service.members()
        .delete(groupKey=google_group_email, memberKey=user_email)
        .execute()
    )


async def remove_user_from_group(
    user_email: str,
    google_group_email: str,
    updated_by: Optional[str] = None,
    dry_run: None = None,
    service: Optional[Resource] = None,
) -> Dict[str, bool]:
    """Remove user member to group."""
    result = {"done": True}
    dynamo = UserDynamoHandler()
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    stats.count(function)
    log_data = {
        "function": function,
        "user_email": user_email,
        "group": google_group_email,
        "updated_by": updated_by,
        "dry_run": dry_run,
        "message": "Removing user from group",
    }
    log.debug(log_data)

    group_info = await auth.get_group_info(google_group_email, members=False)
    await raise_if_restricted(user_email, group_info)
    if not service:
        service = await get_service("admin", "directory_v1", google_group_email)
    existing = await list_group_members(google_group_email, dry_run=dry_run)

    if user_email in existing:
        if not dry_run:
            stats.count(
                f"{function}.remove_user_from_group",
                tags={
                    "user_email": user_email,
                    "google_group_email": google_group_email,
                    "updated_by": updated_by,
                },
            )
            await delete_group_members_call(service, google_group_email, user_email)
            await dynamo.create_group_log_entry(
                google_group_email, user_email, updated_by, "Removed"
            )
    else:
        log_data[
            "message"
        ] = "Unable to remove user from group. User is not currently in the group."
        log.warning(log_data)
        result["done"] = False
        result["message"] = log_data["message"]
        raise NotAMemberException(result["message"])
    return result



import sys
import time
from typing import List

import ujson as json

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    InvalidCertificateException,
    MissingConfigurationValue,
    NoUserException,
)
from consoleme.lib.generic import str2bool
from consoleme.lib.plugins import get_plugin_by_name

log = config.get_logger("consoleme")
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


class Group(object):
    def __init__(self, **kwargs):
        self.name: str = kwargs.get("name")
        self.domain: str = kwargs.get("domain")
        self.group_id: str = kwargs.get("groupId")
        self.friendly_name: str = kwargs.get("friendlyName")
        self.description: str = kwargs.get("description")
        self.settings: str = kwargs.get("settings")
        self.aliases: str = kwargs.get("aliases")
        self.members: List = kwargs.get("members", [])
        self.attributes: List = kwargs.get("attributes")
        self.automated_group: bool = self.is_group_automated(self.description)

        # Set all boolean attributes
        for attr in config.get("groups.attributes.boolean", []):
            attribute_name = attr.get("name")
            setattr(self, attribute_name, str2bool(kwargs.get(attribute_name)))

        # Set all list attributes
        for attr in config.get("groups.attributes.list", []):
            attribute_name = attr.get("name")
            setattr(self, attribute_name, self.convert_to_list(kwargs, attribute_name))

    def get(self, query, default=None):
        result = self.__dict__.get(query)
        if result is None:
            result = default
        return result

    @staticmethod
    def convert_to_list(kwargs, kwarg):
        if kwargs.get(kwarg):
            return kwargs[kwarg].split(",")
        return []

    @staticmethod
    def is_group_automated(description):
        return False


class User(object):
    def __init__(self, **kwargs):
        self.username: str = kwargs.get("userName")
        self.domain: str = kwargs.get("domain")
        self.fullname: str = kwargs.get("name", {}).get("fullName")
        self.status: str = kwargs.get("status")
        self.created: str = kwargs.get("created", {}).get("onDate")
        self.updated: str = kwargs.get("updated", {}).get("onDate")
        self.groups: str = kwargs.get("members")
        self.passed_background_check: bool = str2bool(
            kwargs.get("passed_background_check", False)
        )

    def get(self, query):
        return self.__dict__.get(query)

    def to_json(self):
        d = {
            "username": self.username,
            "domain": self.domain,
            "fullname": self.fullname,
            "status": self.status,
            "created": self.created,
            "updated": self.updated,
            "groups": self.groups,
            "passed_background_check": self.passed_background_check,
        }
        return json.dumps(d)


class ExtendedAttribute(object):
    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.type = kwargs.get("attributeType", "string")
        self.value = kwargs.get("attributeValue")
        if self.value is True:
            self.value = "true"
        elif self.value is False:
            self.value = "false"
        self.sensitive = kwargs.get("sensitive", False)
        self.immutable = kwargs.get("immutable", False)

    def get(self, query):
        return self.__dict__.get(query)

    def to_json(self):
        if isinstance(self.value, list):
            self.value = ",".join(self.value)
        d = {
            "name": self.name,
            "attributeType": self.type,
            "attributeValue": self.value,
            "sensitive": self.sensitive,
            "immutable": self.immutable,
        }
        return json.dumps(d)


class Auth:
    """The Auth class authenticates the user and provides the user's groups."""

    def __init__(self, headers: dict = None):
        """Initialize the auth class."""
        self.headers = headers

    async def get_user(self, headers: dict = None):
        """Get the user identity."""
        if config.get("auth.get_user_by_header"):
            return await self.get_user_by_header(headers)
        else:
            raise Exception("auth.get_user not configured")

    async def get_user_by_header(self, headers: dict):
        """Get the user identity via plaintext header."""
        if not headers:
            raise Exception(
                "auth.get_user_by_header enabled, but no headers were passed in"
            )

        user_header_name = config.get("auth.user_header_name")
        if not user_header_name:
            raise Exception(
                "auth.user_header_name configuration not set, but auth.get_user_by_header is enabled."
            )

        user = headers.get(user_header_name)
        if not user:
            raise NoUserException("User header '{}' is empty.".format(user_header_name))
        return user

    async def get_groups(
        self, user: str, headers=None, get_header_groups=False, only_direct=True
    ):
        """Get the user's groups."""
        groups_to_add_for_all_users = config.get("auth.groups_to_add_for_all_users", [])
        groups = []
        if get_header_groups or config.get("auth.get_groups_by_header"):
            header_groups = await self.get_groups_by_header(headers)
            if header_groups:
                groups.extend(header_groups)
        elif config.get("auth.get_groups_from_google"):
            from consoleme.lib.google import get_group_memberships

            google_groups = await get_group_memberships(user)
            if google_groups:
                groups.extend(google_groups)
        if groups_to_add_for_all_users:
            # Optionally consider ConsoleMe users a member of these additional groups
            groups.extend(groups_to_add_for_all_users)
        if not groups:
            log.error(
                {
                    "message": "auth.get_groups not configured properly or no groups were obtained."
                },
                exc_info=True,
            )
        if config.get("auth.force_groups_lowercase", False):
            groups = [x.lower() for x in groups]
        return list(set(groups))

    async def get_groups_by_header(self, headers: dict):
        """Get the user's groups by plaintext header."""
        groups = []

        if not headers:
            log_data = {
                "function": f"{__name__}.{sys._getframe().f_code.co_name}",
                "message": "No headers present.",
            }
            log.debug(log_data, exc_info=True)
            return groups

        groups_header_name = config.get("auth.groups_header_name", None)
        if not groups_header_name:
            log_data = {
                "function": f"{__name__}.{sys._getframe().f_code.co_name}",
                "message": "Group header name not configured.",
            }
            log.debug(log_data, exc_info=True)
            raise Exception(
                "auth.groups_header_name configuration not set, but auth.get_groups_by_header is enabled."
            )

        if headers.get(groups_header_name):
            groups = headers[groups_header_name].split(",")

        return groups

    async def extract_certificate(self, headers: dict):
        raise NotImplementedError()

    async def extract_user_from_certificate(self, headers: dict):
        return await self.extract_certificate(headers)

    async def get_cert_age_seconds(self, headers: dict):
        """Retrieve age of mtls certificate."""
        current_time = int(time.time())
        cert = await self.extract_certificate(headers)
        return current_time - int(cert.get("notBefore"))

    async def validate_certificate(self, headers: dict):
        cli_auth_required_headers = config.get("cli_auth.required_headers")
        if not cli_auth_required_headers:
            raise MissingConfigurationValue(
                "You must specified the header key and expected value in order to validate a certificate for mutual "
                "TLS authentication. Refer to the `cli_auth.required_headers` configuration"
            )
        for header in cli_auth_required_headers:
            for k, v in header.items():
                if headers.get(k) != v:
                    stats.count("auth.validate_certificate.error")
                    error = (
                        "Header {} is supposed to equal {}, but it equals {}.".format(
                            k, v, headers.get(k)
                        )
                    )
                    log_data = {
                        "function": "auth.validate_certificate",
                        "message": error,
                    }
                    log.error(log_data)
                    raise InvalidCertificateException(error)
        return True

    async def is_user_contractor(self, user):
        return False

    async def validate_and_return_api_caller(self, headers: dict):
        cli_auth_required_headers = config.get("cli_auth.required_headers")
        if not cli_auth_required_headers:
            raise MissingConfigurationValue(
                "You must specified the header key and expected value in order to validate a certificate for mutual "
                "TLS authentication. Refer to the `cli_auth.required_headers` configuration"
            )
        for header in cli_auth_required_headers:
            for k, v in header.items():
                if headers.get(k) != v:
                    raise Exception(
                        f"Header {k} is supposed to equal {v}, but it equals {headers.get(k)}."
                    )
        cert = await self.extract_user_from_certificate(headers)
        user = cert.get("name")
        if not user or user not in config.get("api_auth.valid_entities", []):
            raise Exception("Not authorized to call this API with that certificate.")
        return user

    async def get_user_info(self, user: str, object: bool = False):
        """
        Retrieve details about a user from an authorative source
        :param user:
        :param object:
        :return:
        """
        return {
            "domain": "",
            "userName": user,
            "name": {
                "givenName": "",
                "familyName": "",
                "fullName": "",
            },
            "primaryEmail": user,
        }

    async def get_group_info(self, group, members=True):
        raise NotImplementedError()

    async def put_group_attribute(self, group, attribute_name, attribute_value):
        raise NotImplementedError()

    async def put_group_attributes(self, group, attributes):
        raise NotImplementedError()

    async def is_user_in_group(self, user, group, only_direct=True):
        raise NotImplementedError()

    async def is_requestable(self, group):
        raise NotImplementedError()

    async def does_user_exist(self, user):
        raise NotImplementedError()

    async def get_group_attribute(self, group, attribute_name):
        raise NotImplementedError()

    async def get_secondary_approvers(self, group):
        """Return a list of secondary approvers for a group."""
        raise NotImplementedError()

    async def get_groups_with_attribute_name_value(
        self, attribute_name, attribute_value
    ):
        raise NotImplementedError()

    async def get_users_with_attribute_name_value(
        self, attribute_name, attribute_value
    ):
        raise NotImplementedError()

    async def is_group_requestable(self, group):
        raise NotImplementedError()

    async def get_all_requestable_groups(self):
        raise NotImplementedError()

    async def get_group_memberships(self, user, scopes=[], only_direct=True) -> list:
        return []

    async def get_group_members(self, group):
        raise NotImplementedError()

    async def get_user_attribute(self, user, attribute_name):
        raise NotImplementedError()

    async def get_or_create_user_role_name(self, user):
        user_role_name_attribute = await self.get_user_attribute(user, "user_role_name")
        if not user_role_name_attribute:
            return await self.generate_and_store_user_role_name(user)
        user_role_name = user_role_name_attribute.value
        return user_role_name

    async def generate_and_store_user_role_name(self, user):
        (username, domain) = user.split("@")
        return f"{username}-{domain}"


Auth.Group = Group
Auth.User = User
Auth.ExtendedAttribute = ExtendedAttribute


def init():
    """Initialize the auth plugin."""
    return Auth()



import re
import sys
import time
from datetime import datetime
from typing import Any, Dict

import sentry_sdk
import ujson as json
from asgiref.sync import sync_to_async
from cloudaux.aws.sts import boto3_cached_conn

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    DataNotRetrievable,
    MissingConfigurationValue,
)
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.plugins import get_plugin_by_name

aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
log = config.get_logger()


async def get_resource_from_cloudtrail_deny(ct_event, raw_ct_event):
    """
    Naive attempt to determine resource from Access Deny CloudTrail event. If we can't parse it from the
    Cloudtrail message, we return `*`.
    """

    resources = [
        resource["ARN"]
        for resource in raw_ct_event.get("resources", [])
        if "ARN" in resource
    ]
    if resources:
        resource: str = max(resources, key=len)
        return resource
    resource = "*"

    error_message = ct_event.get("error_message", "")
    if not error_message:
        return resource

    if "on resource: arn:aws" in ct_event.get("error_message", ""):
        resource_re = re.match(
            r"^.* on resource: (arn:aws:.*?$)", ct_event["error_message"]
        )
        if resource_re and len(resource_re.groups()) == 1:
            resource = resource_re.groups()[0]
    return resource


async def generate_policy_from_cloudtrail_deny(ct_event):
    """
    Naive generation of policy from a cloudtrail deny event

    TODOS:
        * Check if resource can already perform action. If so. Check if resource policy permits the action
        * For an action with a resource, check if that resource exists (Assume role to an IAM role, for example)
        * Check if being denied for some other reason than policy allowance. IE SCP, permission boundary, etc.
    """
    if ct_event["error_code"] not in [
        "UnauthorizedOperation",
        "AccessDenied",
    ]:
        return None

    policy = {
        "Statement": [
            {
                "Action": [ct_event["event_call"]],
                "Effect": "Allow",
                "Resource": [ct_event["resource"]],
            }
        ],
        "Version": "2012-10-17",
    }
    return policy


async def detect_cloudtrail_denies_and_update_cache(
    celery_app,
    event_ttl=config.get(
        "event_bridge.detect_cloudtrail_denies_and_update_cache.event_ttl", 86400
    ),
    max_num_messages_to_process=config.get(
        "event_bridge.detect_cloudtrail_denies_and_update_cache.max_num_messages_to_process",
        100,
    ),
) -> Dict[str, Any]:
    log_data = {"function": f"{__name__}.{sys._getframe().f_code.co_name}"}
    dynamo = UserDynamoHandler()
    queue_arn = config.get(
        "event_bridge.detect_cloudtrail_denies_and_update_cache.queue_arn", ""
    ).format(region=config.region)
    if not queue_arn:
        raise MissingConfigurationValue(
            "Unable to find required configuration value: "
            "`event_bridge.detect_cloudtrail_denies_and_update_cache.queue_arn`"
        )
    queue_name = queue_arn.split(":")[-1]
    queue_account_number = queue_arn.split(":")[4]
    queue_region = queue_arn.split(":")[3]
    # Optionally assume a role before receiving messages from the queue
    queue_assume_role = config.get(
        "event_bridge.detect_cloudtrail_denies_and_update_cache.assume_role"
    )

    # Modify existing cloudtrail deny samples
    all_cloudtrail_denies_l = await dynamo.parallel_scan_table_async(
        dynamo.cloudtrail_table
    )
    all_cloudtrail_denies = {}
    for cloudtrail_deny in all_cloudtrail_denies_l:
        all_cloudtrail_denies[cloudtrail_deny["request_id"]] = cloudtrail_deny

    sqs_client = await sync_to_async(boto3_cached_conn)(
        "sqs",
        service_type="client",
        region=queue_region,
        retry_max_attempts=2,
        account_number=queue_account_number,
        assume_role=queue_assume_role,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )

    queue_url_res = await sync_to_async(sqs_client.get_queue_url)(QueueName=queue_name)
    queue_url = queue_url_res.get("QueueUrl")
    if not queue_url:
        raise DataNotRetrievable(f"Unable to retrieve Queue URL for {queue_arn}")
    messages_awaitable = await sync_to_async(sqs_client.receive_message)(
        QueueUrl=queue_url, MaxNumberOfMessages=10
    )
    new_events = 0
    messages = messages_awaitable.get("Messages", [])
    num_events = 0
    reached_limit_on_num_messages_to_process = False

    while messages:
        if num_events >= max_num_messages_to_process:
            reached_limit_on_num_messages_to_process = True
            break
        processed_messages = []
        for message in messages:
            try:
                message_body = json.loads(message["Body"])
                try:
                    if "Message" in message_body:
                        decoded_message = json.loads(message_body["Message"])["detail"]
                    else:
                        decoded_message = message_body["detail"]
                except Exception as e:
                    log.error(
                        {
                            **log_data,
                            "message": "Unable to process Cloudtrail message",
                            "message_body": message_body,
                            "error": str(e),
                        }
                    )
                    sentry_sdk.capture_exception()
                    continue
                event_name = decoded_message.get("eventName")
                event_source = decoded_message.get("eventSource")
                for event_source_substitution in config.get(
                    "event_bridge.detect_cloudtrail_denies_and_update_cache.event_bridge_substitutions",
                    [".amazonaws.com"],
                ):
                    event_source = event_source.replace(event_source_substitution, "")
                event_time = decoded_message.get("eventTime")
                utc_time = datetime.strptime(event_time, "%Y-%m-%dT%H:%M:%SZ")
                epoch_event_time = int(
                    (utc_time - datetime(1970, 1, 1)).total_seconds()
                )
                # Skip entries older than a day
                if int(time.time()) - 86400 > epoch_event_time:
                    continue
                try:
                    session_name = decoded_message["userIdentity"]["arn"].split("/")[-1]
                except (
                    IndexError,
                    KeyError,
                ):  # If IAM user, there won't be a session name
                    session_name = ""
                try:
                    principal_arn = decoded_message["userIdentity"]["sessionContext"][
                        "sessionIssuer"
                    ]["arn"]
                except KeyError:  # Skip events without a parsable ARN
                    continue

                event_call = f"{event_source}:{event_name}"

                ct_event = dict(
                    error_code=decoded_message.get("errorCode"),
                    error_message=decoded_message.get("errorMessage"),
                    arn=principal_arn,
                    # principal_owner=owner,
                    session_name=session_name,
                    source_ip=decoded_message["sourceIPAddress"],
                    event_call=event_call,
                    epoch_event_time=epoch_event_time,
                    ttl=epoch_event_time + event_ttl,
                    count=1,
                )
                resource = await get_resource_from_cloudtrail_deny(
                    ct_event, decoded_message
                )
                ct_event["resource"] = resource
                request_id = f"{principal_arn}-{session_name}-{event_call}-{resource}"
                ct_event["request_id"] = request_id
                generated_policy = await generate_policy_from_cloudtrail_deny(ct_event)
                if generated_policy:
                    ct_event["generated_policy"] = generated_policy

                if all_cloudtrail_denies.get(request_id):
                    existing_count = all_cloudtrail_denies[request_id].get("count", 1)
                    ct_event["count"] += existing_count
                    all_cloudtrail_denies[request_id] = ct_event
                else:
                    all_cloudtrail_denies[request_id] = ct_event
                    new_events += 1
                num_events += 1
            except Exception as e:
                log.error({**log_data, "error": str(e)}, exc_info=True)
                sentry_sdk.capture_exception()
            processed_messages.append(
                {
                    "Id": message["MessageId"],
                    "ReceiptHandle": message["ReceiptHandle"],
                }
            )
        if processed_messages:
            await sync_to_async(sqs_client.delete_message_batch)(
                QueueUrl=queue_url, Entries=processed_messages
            )

        await sync_to_async(dynamo.batch_write_cloudtrail_events)(
            all_cloudtrail_denies.values()
        )
        messages_awaitable = await sync_to_async(sqs_client.receive_message)(
            QueueUrl=queue_url, MaxNumberOfMessages=10
        )
        messages = messages_awaitable.get("Messages", [])
    if reached_limit_on_num_messages_to_process:
        # We hit our limit. Let's spawn another task immediately to process remaining messages
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_cloudtrail_denies",
        )
    log_data["message"] = "Successfully cached Cloudtrail Access Denies"
    log_data["num_events"] = num_events
    log_data["new_events"] = new_events
    log.debug(log_data)

    return log_data



import boto3

from consoleme.config import config
from consoleme.models import CloudAccountModel, CloudAccountModelArray


async def retrieve_current_account() -> CloudAccountModelArray:
    client = boto3.client("sts", **config.get("boto3.client_kwargs", {}))
    identity = client.get_caller_identity()
    account_aliases = boto3.client(
        "iam", **config.get("boto3.client_kwargs", {})
    ).list_account_aliases()["AccountAliases"]
    account_id = None
    if identity and identity.get("Account"):
        account_id = identity.get("Account")
    account_name = account_id

    if account_aliases:
        account_name = account_aliases[0]

    cloud_account = [
        CloudAccountModel(
            id=account_id,
            name=account_name,
            status="active",
            sync_enabled=True,
            type="aws",
        )
    ]
    return CloudAccountModelArray(accounts=cloud_account)



from pydantic import BaseModel as PydanticBaseModel


def to_camel(string):
    """Convert a snake_case string to CamelCase"""
    return "".join(word.capitalize() for word in string.split("_"))


class BaseModel(PydanticBaseModel):
    """BaseModel adds CamelCase aliases to a Pydantic model"""

    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True



from datetime import datetime, timedelta
from typing import List, Optional, Union

import ujson as json
from asgiref.sync import sync_to_async
from policy_sentry.util.arns import parse_arn

from consoleme.config import config
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.crypto import Crypto
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.policies import get_aws_config_history_url_for_resource
from consoleme.lib.redis import RedisHandler, redis_get
from consoleme.models import (
    AwsPrincipalModel,
    CloudTrailDetailsModel,
    CloudTrailError,
    CloudTrailErrorArray,
    EligibleRolesModel,
    EligibleRolesModelArray,
    ExtendedAwsPrincipalModel,
    S3DetailsModel,
    S3Error,
    S3ErrorArray,
)

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()
crypto = Crypto()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
internal_policies = get_plugin_by_name(
    config.get("plugins.internal_policies", "default_policies")
)()
red = RedisHandler().redis_sync()


async def get_config_timeline_url_for_role(role, account_id):
    resource_id = role.get("resourceId")
    if resource_id:
        config_history_url = await get_aws_config_history_url_for_resource(
            account_id, resource_id, role["arn"], "AWS::IAM::Role"
        )
        return config_history_url


async def get_cloudtrail_details_for_role(arn: str):
    """
    Retrieves CT details associated with role, if they exist exists
    :param arn:
    :return:
    """
    error_url = config.get("cloudtrail_errors.error_messages_by_role_uri", "").format(
        arn=arn
    )

    errors_unformatted = await internal_policies.get_errors_by_role(
        arn, config.get("policies.number_cloudtrail_errors_to_display", 5)
    )

    ct_errors = []

    for event_string, value in errors_unformatted.items():
        event_call = event_string
        resource = None
        generated_policy = value.get("generated_policy", None)
        if "|||" in event_string:
            event_call, resource = event_string.split("|||")

        ct_errors.append(
            CloudTrailError(
                event_call=event_call,
                resource=resource,
                generated_policy=generated_policy,
                count=value.get("count", 0),
            )
        )

    return CloudTrailDetailsModel(
        error_url=error_url, errors=CloudTrailErrorArray(cloudtrail_errors=ct_errors)
    )


async def get_s3_details_for_role(account_id: str, role_name: str) -> S3DetailsModel:
    """
    Retrieves s3 details associated with role, if it exists
    :param arn:
    :return:
    """
    arn = f"arn:aws:iam::{account_id}:role/{role_name}"
    yesterday = (datetime.today() - timedelta(days=1)).strftime("%Y%m%d")
    error_url = config.get("s3.query_url", "").format(
        yesterday=yesterday, role_name=f"'{role_name}'", account_id=f"'{account_id}'"
    )
    query_url = config.get("s3.non_error_query_url", "").format(
        yesterday=yesterday, role_name=f"'{role_name}'", account_id=f"'{account_id}'"
    )

    s3_error_topic = config.get("redis.s3_errors", "S3_ERRORS")
    all_s3_errors = await redis_get(s3_error_topic)
    s3_errors_unformatted = []
    if all_s3_errors:
        s3_errors_unformatted = json.loads(all_s3_errors).get(arn, [])
    s3_errors_formatted = []
    for error in s3_errors_unformatted:
        s3_errors_formatted.append(
            S3Error(
                count=error.get("count", ""),
                bucket_name=error.get("bucket_name", ""),
                request_prefix=error.get("request_prefix", ""),
                error_call=error.get("error_call", ""),
                status_code=error.get("status_code", ""),
                status_text=error.get("status_text", ""),
                role_arn=arn,
            )
        )

    return S3DetailsModel(
        query_url=query_url,
        error_url=error_url,
        errors=S3ErrorArray(s3_errors=s3_errors_formatted),
    )


async def get_app_details_for_role(arn: str):
    """
    Retrieves applications associated with role, if they exist
    :param arn:
    :return:
    """
    return await internal_policies.get_applications_associated_with_role(arn)


async def get_role_template(arn: str):
    return await sync_to_async(red.hget)(
        config.get("templated_roles.redis_key", "TEMPLATED_ROLES_v2"), arn.lower()
    )


async def get_user_details(
    account_id: str, user_name: str, extended: bool = False, force_refresh: bool = False
) -> Optional[Union[ExtendedAwsPrincipalModel, AwsPrincipalModel]]:
    account_ids_to_name = await get_account_id_to_name_mapping()
    arn = f"arn:aws:iam::{account_id}:user/{user_name}"
    user = await aws.fetch_iam_user(account_id, arn)
    # requested user doesn't exist
    if not user:
        return None
    if extended:
        return ExtendedAwsPrincipalModel(
            name=user_name,
            account_id=account_id,
            account_name=account_ids_to_name.get(account_id, None),
            arn=arn,
            inline_policies=user.get("UserPolicyList", []),
            config_timeline_url=await get_config_timeline_url_for_role(
                user, account_id
            ),
            cloudtrail_details=await get_cloudtrail_details_for_role(arn),
            s3_details=await get_s3_details_for_role(
                account_id=account_id, role_name=user_name
            ),
            apps=await get_app_details_for_role(arn),
            managed_policies=user["AttachedManagedPolicies"],
            groups=user["Groups"],
            tags=user["Tags"],
            owner=user.get("owner"),
            templated=False,
            template_link=None,
            created_time=str(user.get("CreateDate", "")),
            last_used_time=user.get("RoleLastUsed", {}).get("LastUsedDate"),
            description=user.get("Description"),
            permissions_boundary=user.get("PermissionsBoundary", {}),
        )
    else:
        return AwsPrincipalModel(
            name=user_name,
            account_id=account_id,
            account_name=account_ids_to_name.get(account_id, None),
            arn=arn,
        )


async def get_role_details(
    account_id: str, role_name: str, extended: bool = False, force_refresh: bool = False
) -> Optional[Union[ExtendedAwsPrincipalModel, AwsPrincipalModel]]:
    account_ids_to_name = await get_account_id_to_name_mapping()
    arn = f"arn:aws:iam::{account_id}:role/{role_name}"
    role = await aws.fetch_iam_role(account_id, arn, force_refresh=force_refresh)
    # requested role doesn't exist
    if not role:
        return None
    if extended:
        template = await get_role_template(arn)
        return ExtendedAwsPrincipalModel(
            name=role_name,
            account_id=account_id,
            account_name=account_ids_to_name.get(account_id, None),
            arn=arn,
            inline_policies=role["policy"].get(
                "RolePolicyList", role["policy"].get("UserPolicyList", [])
            ),
            assume_role_policy_document=role["policy"]["AssumeRolePolicyDocument"],
            config_timeline_url=await get_config_timeline_url_for_role(
                role, account_id
            ),
            cloudtrail_details=await get_cloudtrail_details_for_role(arn),
            s3_details=await get_s3_details_for_role(
                account_id=account_id, role_name=role_name
            ),
            apps=await get_app_details_for_role(arn),
            managed_policies=role["policy"]["AttachedManagedPolicies"],
            tags=role["policy"]["Tags"],
            templated=bool(template),
            template_link=template,
            created_time=role["policy"].get("CreateDate"),
            last_used_time=role["policy"].get("RoleLastUsed", {}).get("LastUsedDate"),
            description=role["policy"].get("Description"),
            owner=role.get("owner"),
            permissions_boundary=role["policy"].get("PermissionsBoundary", {}),
        )
    else:
        return AwsPrincipalModel(
            name=role_name,
            account_id=account_id,
            account_name=account_ids_to_name.get(account_id, None),
            arn=arn,
        )


async def get_eligible_role_details(
    eligible_roles: List[str],
) -> EligibleRolesModelArray:
    account_ids_to_name = await get_account_id_to_name_mapping()
    eligible_roles_detailed = []
    for role in eligible_roles:
        arn_parsed = parse_arn(role)
        account_id = arn_parsed["account"]
        role_name = (
            arn_parsed["resource_path"]
            if arn_parsed["resource_path"]
            else arn_parsed["resource"]
        )
        account_friendly_name = account_ids_to_name.get(account_id, "Unknown")
        role_apps = await get_app_details_for_role(role)
        eligible_roles_detailed.append(
            EligibleRolesModel(
                arn=role,
                account_id=account_id,
                account_friendly_name=account_friendly_name,
                role_name=role_name,
                apps=role_apps,
            )
        )

    return EligibleRolesModelArray(roles=eligible_roles_detailed)



from typing import Dict

import ujson as json

from consoleme.config import config
from consoleme.lib.cache import retrieve_json_data_from_redis_or_s3
from consoleme.lib.cloud_credential_authorization_mapping.models import (
    CredentialAuthzMappingGenerator,
    RoleAuthorizations,
    user_or_group,
)


class RoleTagAuthorizationMappingGenerator(CredentialAuthzMappingGenerator):
    """ Generates an authorization mapping of groups -> roles based on IAM role tags."""

    async def generate_credential_authorization_mapping(
        self, authorization_mapping: Dict[user_or_group, RoleAuthorizations]
    ) -> Dict[user_or_group, RoleAuthorizations]:
        """This will list accounts that meet the account attribute search criteria."""
        # Retrieve roles
        cache_key = config.get("aws.iamroles_redis_key", "IAM_ROLE_CACHE")
        all_roles = await retrieve_json_data_from_redis_or_s3(
            redis_key=cache_key,
            redis_data_type="hash",
            s3_bucket=config.get(
                "cache_iam_resources_across_accounts.all_roles_combined.s3.bucket"
            ),
            s3_key=config.get(
                "cache_iam_resources_across_accounts.all_roles_combined.s3.file",
                "account_resource_cache/cache_all_roles_v1.json.gz",
            ),
        )

        required_trust_policy_entity = config.get(
            "cloud_credential_authorization_mapping.role_tags.required_trust_policy_entity"
        )

        for arn, role_entry_j in all_roles.items():
            role_entry = json.loads(role_entry_j)
            policy = json.loads(role_entry["policy"])
            tags = policy.get("Tags")

            if (
                required_trust_policy_entity
                and required_trust_policy_entity.lower()
                not in json.dumps(
                    policy["AssumeRolePolicyDocument"], escape_forward_slashes=False
                ).lower()
            ):
                continue

            for tag in tags:
                if tag["Key"] in config.get(
                    "cloud_credential_authorization_mapping.role_tags.authorized_groups_tags",
                    [],
                ):
                    splitted_groups = tag["Value"].split(":")
                    for group in splitted_groups:
                        if config.get("auth.force_groups_lowercase", False):
                            group = group.lower()
                        if not authorization_mapping.get(group):
                            authorization_mapping[group] = RoleAuthorizations.parse_obj(
                                {
                                    "authorized_roles": set(),
                                    "authorized_roles_cli_only": set(),
                                }
                            )
                        authorization_mapping[group].authorized_roles.add(arn)
                if tag["Key"] in config.get(
                    "cloud_credential_authorization_mapping.role_tags.authorized_groups_cli_only_tags",
                    [],
                ):
                    splitted_groups = tag["Value"].split(":")
                    for group in splitted_groups:
                        if config.get("auth.force_groups_lowercase", False):
                            group = group.lower()
                        if not authorization_mapping.get(group):
                            authorization_mapping[group] = RoleAuthorizations.parse_obj(
                                {
                                    "authorized_roles": set(),
                                    "authorized_roles_cli_only": set(),
                                }
                            )
                        authorization_mapping[group].authorized_roles_cli_only.add(arn)
        return authorization_mapping



import fnmatch
import os
import sys
import tempfile
from typing import Optional, Union

import sentry_sdk
from ruamel.yaml import YAML

from consoleme.config import config
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.cache import (
    retrieve_json_data_from_redis_or_s3,
    store_json_results_in_redis_and_s3,
)
from consoleme.lib.git import clone_repo
from consoleme.lib.templated_resources.models import (
    TemplatedFileModelArray,
    TemplateFile,
)

log = config.get_logger()

yaml = YAML()
yaml.preserve_quotes = True
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.width = 4096


async def cache_resource_templates() -> TemplatedFileModelArray:
    templated_file_array = TemplatedFileModelArray(templated_resources=[])
    for repository in config.get("cache_resource_templates.repositories", []):
        if repository.get("type") == "git":
            result = await cache_resource_templates_for_repository(repository)
            templated_file_array.templated_resources.extend(result.templated_resources)
    await store_json_results_in_redis_and_s3(
        templated_file_array.dict(),
        redis_key=config.get(
            "cache_resource_templates.redis.key", "cache_templated_resources_v1"
        ),
        s3_bucket=config.get("cache_resource_templates.s3.bucket"),
        s3_key=config.get(
            "cache_resource_templates.s3.file",
            "cache_templated_resources/cache_templated_resources_v1.json.gz",
        ),
    )
    return templated_file_array


async def retrieve_cached_resource_templates(
    resource_type: Optional[str] = None,
    resource: Optional[str] = None,
    repository_name: Optional[str] = None,
    template_language: Optional[str] = None,
    return_first_result=False,
) -> Optional[Union[TemplatedFileModelArray, TemplateFile]]:
    matching_templates = []
    templated_resource_data_d = await retrieve_json_data_from_redis_or_s3(
        redis_key=config.get(
            "cache_resource_templates.redis.key", "cache_templated_resources_v1"
        ),
        s3_bucket=config.get("cache_resource_templates.s3.bucket"),
        s3_key=config.get(
            "cache_resource_templates.s3.file",
            "cache_templated_resources/cache_templated_resources_v1.json.gz",
        ),
    )
    templated_file_array = TemplatedFileModelArray.parse_obj(templated_resource_data_d)
    for template_file in templated_file_array.templated_resources:
        if resource_type and not template_file.resource_type == resource_type:
            continue
        if resource and not template_file.resource == resource:
            continue
        if repository_name and not template_file.repository_name == repository_name:
            continue
        if (
            template_language
            and not template_file.template_language == template_language
        ):
            continue
        if return_first_result:
            return template_file
        matching_templates.append(template_file)
    if return_first_result:
        return None
    return TemplatedFileModelArray(templated_resources=matching_templates)


async def cache_resource_templates_for_repository(
    repository,
) -> TemplatedFileModelArray:
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    log_data = {
        "function": function,
        "repository": repository,
    }
    if repository["type"] not in ["git"]:
        raise Exception("Unsupported repository type")
    tempdir = tempfile.mkdtemp()
    repo_url = repository["repo_url"]
    repo = clone_repo(repo_url, tempdir)
    repo.config_writer().set_value("user", "name", "ConsoleMe").release()
    email = repository["authentication_settings"]["email"]
    resource_formats = repository["resource_formats"]
    discovered_templates = []
    accounts_d = await get_account_id_to_name_mapping()
    accounts_set = set(accounts_d.values())
    if email:
        repo.config_writer().set_value("user", "email", email).release()
    for subdir, dirs, files in os.walk(repo.working_dir):
        for filename in files:
            template_matched = False
            subdir = subdir.replace(tempdir + os.sep, "")
            filepath = subdir + os.sep + filename
            relative_path = "/".join(subdir.split("/")[1:]) + os.sep + filename
            web_path = repository.get("web_path", "").format(
                relative_path=relative_path
            )
            full_temporary_file_path = tempdir + "/" + filepath
            if "honeybee" in resource_formats:
                for resource_type, conditions in repository["resource_type_parser"][
                    "honeybee"
                ].items():
                    if template_matched:
                        continue
                    for condition in conditions:
                        if condition.get("path_prefix") and not filepath.startswith(
                            condition["path_prefix"]
                        ):
                            continue
                        if condition.get("path_suffix") and not filepath.endswith(
                            condition["path_suffix"]
                        ):
                            continue
                        with open(full_temporary_file_path, "r") as f:
                            try:
                                file_content = yaml.load(f)
                            except Exception as e:
                                sentry_sdk.capture_exception()
                                log.error(
                                    {
                                        **log_data,
                                        "Message": "Error trying to parse template",
                                        "file_path": full_temporary_file_path,
                                        "error": str(e),
                                    },
                                    exc_info=True,
                                )
                                continue
                        name = file_content.get(
                            "TemplateName", file_content.get("Name", filename)
                        )
                        owner = file_content.get("Owner")

                        # Generate a set of accounts the template applies to. This is used to get the number of accounts
                        # affected by a resource template.
                        included_accounts_set = set()
                        include_accounts = file_content.get("IncludeAccounts", [])
                        if include_accounts:
                            for include_account in include_accounts:
                                for account in accounts_set:
                                    if fnmatch.fnmatch(account, include_account):
                                        included_accounts_set.add(account)
                        exclude_accounts = file_content.get("ExcludeAccounts", [])
                        if exclude_accounts:
                            for exclude_account in exclude_accounts:
                                for account in accounts_set:
                                    if fnmatch.fnmatch(account, exclude_account):
                                        if account not in included_accounts_set:
                                            continue
                                        included_accounts_set.remove(account)

                        if condition.get("file_content"):
                            should_ignore_file = False
                            body_should_include = condition["file_content"].get(
                                "includes", []
                            )
                            body_should_exclude = condition["file_content"].get(
                                "excludes", []
                            )
                            for include in body_should_include:
                                if include not in file_content:
                                    should_ignore_file = True
                                    break
                            for exclude in body_should_exclude:
                                if exclude in file_content:
                                    should_ignore_file = True
                            if should_ignore_file:
                                continue

                        discovered_templates.append(
                            TemplateFile(
                                name=name,
                                repository_name=repository["name"],
                                owner=owner,
                                include_accounts=include_accounts,
                                exclude_accounts=exclude_accounts,
                                number_of_accounts=len(included_accounts_set),
                                resource=relative_path,
                                file_path=filepath,
                                web_path=web_path,
                                resource_type=resource_type,
                                template_language="honeybee",
                            )
                        )
                        template_matched = True
                        break
    return TemplatedFileModelArray(templated_resources=discovered_templates)



from marshmallow import Schema, fields


class ManagedPolicy(Schema):
    action = fields.Str(required=True)
    arn = fields.Str(required=True)


class InlinePolicy(Schema):
    action = fields.Str(required=True)
    policy_name = fields.Str(required=True, load_from="policyname")
    policy_document = fields.Str(load_from="policydocument")


class Tags(Schema):
    action = fields.Str(required=True)
    key = fields.Str(required=True)
    value = fields.Str()


class AssumeRolePolicyDocument(Schema):
    action = fields.Str(required=True)
    assume_role_policy_document = fields.Str(required=True)


class RoleUpdaterRequest(Schema):
    arn = fields.Str(required=True)
    inline_policies = fields.Nested(InlinePolicy, many=True)
    managed_policies = fields.Nested(ManagedPolicy, many=True)
    assume_role_policy_document = fields.Nested(AssumeRolePolicyDocument)
    tags = fields.Nested(Tags, many=True)
    requester = fields.Str()



import base64
import sys
from urllib.parse import parse_qs, urlencode, urlparse

import sentry_sdk
import tornado.escape
import ujson as json
from furl import furl
from pydantic import ValidationError

from consoleme.celery_tasks.celery_tasks import app as celery_app
from consoleme.config import config
from consoleme.handlers.base import BaseAPIV2Handler, BaseMtlsHandler
from consoleme.lib.auth import (
    can_create_roles,
    can_delete_iam_principals,
    can_delete_iam_principals_app,
)
from consoleme.lib.aws import (
    allowed_to_sync_role,
    clone_iam_role,
    create_iam_role,
    delete_iam_role,
)
from consoleme.lib.crypto import Crypto
from consoleme.lib.generic import str2bool
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.v2.aws_principals import get_eligible_role_details, get_role_details
from consoleme.models import (
    CloneRoleRequestModel,
    RoleCreationRequestModel,
    Status2,
    WebResponse,
)

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()
crypto = Crypto()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
group_mapping = get_plugin_by_name(
    config.get("plugins.group_mapping", "default_group_mapping")
)()
internal_policies = get_plugin_by_name(
    config.get("plugins.internal_policies", "default_policies")
)()


class RoleConsoleLoginHandler(BaseAPIV2Handler):
    async def get(self, role=None):
        """
        Attempt to retrieve credentials and redirect the user to the AWS Console
        ---
        description: Retrieves credentials and redirects user to AWS console.
        responses:
            302:
                description: Redirects to AWS console
        """
        arguments = {k: self.get_argument(k) for k in self.request.arguments}
        role = role.lower()
        selected_roles = await group_mapping.filter_eligible_roles(role, self)
        region = arguments.get("r", config.get("aws.region", "us-east-1"))
        redirect = arguments.get("redirect")
        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "role": role,
            "region": region,
            "redirect": redirect,
            "ip": self.ip,
        }

        log_data["role"] = role
        if not selected_roles:
            # Not authorized
            stats.count(
                "RoleConsoleLoginHandler.post",
                tags={
                    "user": self.user,
                    "role": role,
                    "authorized": False,
                    "redirect": bool(redirect),
                },
            )
            log_data[
                "message"
            ] = "You do not have any roles matching your search criteria. "
            log.debug(log_data)
            self.set_status(404)
            self.write({"type": "error", "message": log_data["message"]})
            return

        stats.count(
            "RoleConsoleLoginHandler.post",
            tags={
                "user": self.user,
                "role": role,
                "authorized": True,
                "redirect": bool(redirect),
            },
        )

        if len(selected_roles) > 1:
            # Not sure which role the user wants. Redirect them to main page to select one.
            stats.count(
                "RoleConsoleLoginHandler.post",
                tags={
                    "user": self.user,
                    "role": role,
                    "authorized": False,
                    "redirect": bool(redirect),
                },
            )
            log_data[
                "message"
            ] = "You have more than one role matching your query. Please select one."
            log.debug(log_data)
            warning_message_arg = {
                "warningMessage": base64.b64encode(log_data["message"].encode()).decode(
                    "utf-8"
                )
            }
            redirect_url = furl(f"/?arn={role}")
            redirect_url.args = {
                **redirect_url.args,
                **arguments,
                **warning_message_arg,
            }
            self.write(
                {
                    "type": "redirect",
                    "message": log_data["message"],
                    "reason": "multiple_roles",
                    "redirect_url": redirect_url.url,
                }
            )
            return

        log_data["message"] = "Incoming request"
        log.debug(log_data)

        # User is authorized
        try:
            # User-role logic:
            # User-role should come in as cm-[username or truncated username]_[N or NC]
            user_role = False
            account_id = None

            selected_role = selected_roles[0]

            # User role must be defined as a user attribute
            if (
                self.user_role_name
                and "role/" in selected_role
                and selected_role.split("role/")[1] == self.user_role_name
            ):
                user_role = True
                account_id = selected_role.split("arn:aws:iam::")[1].split(":role")[0]

            url = await aws.generate_url(
                self.user,
                selected_role,
                region,
                user_role=user_role,
                account_id=account_id,
            )
        except Exception as e:
            log_data["message"] = f"Exception generating AWS console URL: {str(e)}"
            log_data["error"] = str(e)
            log.error(log_data, exc_info=True)
            stats.count("index.post.exception")
            self.write(
                {
                    "type": "console_url",
                    "message": tornado.escape.xhtml_escape(log_data["message"]),
                    "error": tornado.escape.xhtml_escape(str(log_data["error"])),
                }
            )
            return
        if redirect:
            parsed_url = urlparse(url)
            parsed_url_query = parse_qs(parsed_url.query)
            parsed_url_query["Destination"] = redirect
            updated_query = urlencode(parsed_url_query, doseq=True)
            url = parsed_url._replace(query=updated_query).geturl()
        self.write(
            {
                "type": "redirect",
                "redirect_url": url,
                "reason": "console_login",
                "role": selected_role,
            }
        )
        return


class RolesHandler(BaseAPIV2Handler):
    """Handler for /api/v2/roles

    GET - Allows read access to a list of roles across all accounts. Returned roles are
    limited to what the requesting user has access to.
    POST - Allows (authorized) users to create a role
    """

    allowed_methods = ["GET", "POST"]

    def on_finish(self) -> None:
        if self.request.method != "POST":
            return
        # Force refresh of crednetial authorization mapping after the dynamic config sync period to ensure all workers
        # have the updated configuration
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_policies_table_details",
        )
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_credential_authorization_mapping",
        )

    async def get(self):
        payload = {"eligible_roles": self.eligible_roles}
        self.set_header("Content-Type", "application/json")
        self.write(json.dumps(payload, escape_forward_slashes=False))
        await self.finish()

    async def post(self):
        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "ip": self.ip,
        }
        can_create_role = can_create_roles(self.user, self.groups)
        if not can_create_role:
            stats.count(
                f"{log_data['function']}.unauthorized",
                tags={"user": self.user, "authorized": can_create_role},
            )
            log_data["message"] = "User is unauthorized to create a role"
            log.error(log_data)
            self.write_error(403, message="User is unauthorized to create a role")
            return

        try:
            create_model = RoleCreationRequestModel.parse_raw(self.request.body)
        except ValidationError as e:
            log_data["message"] = f"Validation Exception: {str(e)}"
            log_data["error"] = str(e)
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.validation_exception", tags={"user": self.user}
            )
            sentry_sdk.capture_exception()
            self.write_error(400, message="Error validating input: " + str(e))
            return

        try:
            results = await create_iam_role(create_model, self.user)
        except Exception as e:
            log_data["message"] = f"Exception creating role: {str(e)}"
            log_data["error"] = str(e)
            log_data["account_id"] = create_model.account_id
            log_data["role_name"] = create_model.role_name
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.exception",
                tags={
                    "user": self.user,
                    "account_id": create_model.account_id,
                    "role_name": create_model.role_name,
                },
            )
            sentry_sdk.capture_exception()
            self.write_error(500, message="Exception occurred cloning role: " + str(e))
            return

        # if here, role has been successfully cloned
        stats.count(
            f"{log_data['function']}.success",
            tags={
                "user": self.user,
                "account_id": create_model.account_id,
                "role_name": create_model.role_name,
            },
        )
        self.write(results)


class AccountRolesHandler(BaseAPIV2Handler):
    """Handler for /api/v2/roles/{account_number}

    Allows read access to a list of roles by account. Roles are limited to what the
    requesting user has access to.
    """

    allowed_methods = ["GET"]

    async def get(self, account_id):
        """
        GET /api/v2/roles/{account_id}
        """
        log_data = {
            "function": "AccountRolesHandler.get",
            "user": self.user,
            "message": "Writing all eligible user roles",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }
        log.debug(log_data)
        self.write_error(501, message="Get roles by account")


class RoleDetailHandler(BaseAPIV2Handler):
    """Handler for /api/v2/roles/{accountNumber}/{roleName}

    Allows read and update access to a specific role in an account.
    """

    allowed_methods = ["GET", "PUT", "DELETE"]

    def initialize(self):
        self.user: str = None
        self.eligible_roles: list = []

    async def get(self, account_id, role_name):
        """
        GET /api/v2/roles/{account_number}/{role_name}
        """

        log_data = {
            "function": "RoleDetailHandler.get",
            "user": self.user,
            "ip": self.ip,
            "message": "Retrieving role details",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "account_id": account_id,
            "role_name": role_name,
        }
        stats.count(
            "RoleDetailHandler.get",
            tags={"user": self.user, "account_id": account_id, "role_name": role_name},
        )
        log.debug(log_data)
        force_refresh = str2bool(
            self.request.arguments.get("force_refresh", [False])[0]
        )

        error = ""

        try:
            role_details = await get_role_details(
                account_id, role_name, extended=True, force_refresh=force_refresh
            )
        except Exception as e:
            sentry_sdk.capture_exception()
            log.error({**log_data, "error": e}, exc_info=True)
            role_details = None
            error = str(e)

        if role_details:
            if not allowed_to_sync_role(role_details.arn, role_details.tags):
                role_details = None

        if not role_details:
            self.send_error(
                404,
                message=f"Unable to retrieve the specified role: {account_id}/{role_name}. {error}",
            )
            return
        self.write(role_details.json())

    async def put(self, account_id, role_name):
        """
        PUT /api/v2/roles/{account_number}/{role_name}
        """
        log_data = {
            "function": "RoleDetailHandler.put",
            "user": self.user,
            "message": "Writing all eligible user roles",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }
        log.debug(log_data)
        self.write_error(501, message="Update role details")

    async def delete(self, account_id, role_name):
        """
        DELETE /api/v2/roles/{account_id}/{role_name}
        """
        if not self.user:
            self.write_error(403, message="No user detected")
            return

        account_id = tornado.escape.xhtml_escape(account_id)
        role_name = tornado.escape.xhtml_escape(role_name)

        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "ip": self.ip,
            "account": account_id,
            "role": role_name,
        }

        can_delete_role = can_delete_iam_principals(self.user, self.groups)
        if not can_delete_role:
            stats.count(
                f"{log_data['function']}.unauthorized",
                tags={
                    "user": self.user,
                    "account": account_id,
                    "role": role_name,
                    "authorized": can_delete_role,
                    "ip": self.ip,
                },
            )
            log_data["message"] = "User is unauthorized to delete a role"
            log.error(log_data)
            self.write_error(403, message="User is unauthorized to delete a role")
            return
        try:
            await delete_iam_role(account_id, role_name, self.user)
        except Exception as e:
            log_data["message"] = "Exception deleting role"
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.exception",
                tags={
                    "user": self.user,
                    "account": account_id,
                    "role": role_name,
                    "authorized": can_delete_role,
                    "ip": self.ip,
                },
            )
            self.write_error(500, message="Error occurred deleting role: " + str(e))
            return

        # if here, role has been successfully deleted
        arn = f"arn:aws:iam::{account_id}:role/{role_name}"
        await aws.fetch_iam_role(account_id, arn, force_refresh=True)
        response_json = {
            "status": "success",
            "message": "Successfully deleted role from account",
            "role": role_name,
            "account": account_id,
        }
        self.write(response_json)


class RoleDetailAppHandler(BaseMtlsHandler):

    """Handler for /api/v2/mtls/roles/{accountNumber}/{roleName}

    Allows apps to view or delete a role
    """

    allowed_methods = ["DELETE", "GET"]

    def check_xsrf_cookie(self):
        pass

    async def delete(self, account_id, role_name):
        """
        DELETE /api/v2/mtls/roles/{account_id}/{role_name}
        """
        account_id = tornado.escape.xhtml_escape(account_id)
        role_name = tornado.escape.xhtml_escape(role_name)
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "account_id": account_id,
            "role_name": role_name,
        }
        requester_type = self.requester.get("type")
        if requester_type != "application":
            log_data[
                "message"
            ] = "Non-application trying to access application only endpoint"
            log.error(log_data)
            self.write_error(406, message="Endpoint not supported for non-applications")
            return

        app_name = self.requester.get("name")
        can_delete_role = can_delete_iam_principals_app(app_name)

        if not can_delete_role:
            stats.count(
                f"{log_data['function']}.unauthorized",
                tags={
                    "app_name": app_name,
                    "account_id": account_id,
                    "role_name": role_name,
                    "authorized": can_delete_role,
                },
            )
            log_data["message"] = "App is unauthorized to delete a role"
            log.error(log_data)
            self.write_error(403, message="App is unauthorized to delete a role")
            return

        try:
            await delete_iam_role(account_id, role_name, app_name)
        except Exception as e:
            log_data["message"] = "Exception deleting role"
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.exception",
                tags={
                    "app_name": app_name,
                    "account_id": account_id,
                    "role_name": role_name,
                    "authorized": can_delete_role,
                },
            )
            self.write_error(500, message="Error occurred deleting role: " + str(e))
            return

        # if here, role has been successfully deleted
        arn = f"arn:aws:iam::{account_id}:role/{role_name}"
        await aws.fetch_iam_role(account_id, arn, force_refresh=True)
        response_json = {
            "status": "success",
            "message": "Successfully deleted role from account",
            "role": role_name,
            "account": account_id,
        }
        self.write(response_json)

    async def get(self, account_id, role_name):
        """
        GET /api/v2/mtls/roles/{account_id}/{role_name}
        """
        account_id = tornado.escape.xhtml_escape(account_id)
        role_name = tornado.escape.xhtml_escape(role_name)
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "ip": self.ip,
            "message": "Retrieving role details",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "account_id": account_id,
            "role_name": role_name,
        }
        app_name = self.requester.get("name") or self.requester.get("username")
        stats.count(
            "RoleDetailAppHandler.get",
            tags={
                "requester": app_name,
                "account_id": account_id,
                "role_name": role_name,
            },
        )
        log.debug(log_data)
        force_refresh = str2bool(
            self.request.arguments.get("force_refresh", [False])[0]
        )

        error = ""

        try:
            role_details = await get_role_details(
                account_id, role_name, extended=True, force_refresh=force_refresh
            )
        except Exception as e:
            sentry_sdk.capture_exception()
            log.error({**log_data, "error": e}, exc_info=True)
            role_details = None
            error = str(e)

        if not role_details:
            self.send_error(
                404,
                message=f"Unable to retrieve the specified role: {account_id}/{role_name}. {error}",
            )
            return
        self.write(role_details.json())


class RoleCloneHandler(BaseAPIV2Handler):
    """Handler for /api/v2/clone/role

    Allows cloning a role.
    """

    allowed_methods = ["POST"]

    async def post(self):
        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "ip": self.ip,
        }
        can_create_role = can_create_roles(self.user, self.groups)
        if not can_create_role:
            stats.count(
                f"{log_data['function']}.unauthorized",
                tags={"user": self.user, "authorized": can_create_role},
            )
            log_data["message"] = "User is unauthorized to clone a role"
            log.error(log_data)
            self.write_error(403, message="User is unauthorized to clone a role")
            return

        try:
            clone_model = CloneRoleRequestModel.parse_raw(self.request.body)
        except ValidationError as e:
            log_data["message"] = "Validation Exception"
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.validation_exception", tags={"user": self.user}
            )
            sentry_sdk.capture_exception()
            self.write_error(400, message="Error validating input: " + str(e))
            return

        try:
            results = await clone_iam_role(clone_model, self.user)
        except Exception as e:
            log_data["message"] = "Exception cloning role"
            log_data["error"] = str(e)
            log_data["account_id"] = clone_model.account_id
            log_data["role_name"] = clone_model.role_name
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.exception",
                tags={
                    "user": self.user,
                    "account_id": clone_model.account_id,
                    "role_name": clone_model.role_name,
                },
            )
            sentry_sdk.capture_exception()
            self.write_error(500, message="Exception occurred cloning role: " + str(e))
            return

        # if here, role has been successfully cloned
        self.write(results)


class GetRolesMTLSHandler(BaseMtlsHandler):
    """
    Handler for /api/v2/get_roles
    Consoleme MTLS role handler - returns User's eligible roles and other details about eligible roles
    Pass ?all=true to URL query to return all roles.
    """

    def check_xsrf_cookie(self):
        pass

    def initialize(self):
        self.user: str = None
        self.eligible_roles: list = []

    async def get(self):
        """
        GET /api/v2/get_roles - Endpoint used to get details of eligible roles. Used by weep and newt.
        ---
        get:
            description: Returns a json-encoded list of objects of eligible roles for the user.
            response format: WebResponse. The "data" field within WebResponse is of format EligibleRolesModelArray
            Example response:
                {
                    "status": "success",
                    "status_code": 200,
                    "data": {
                        "roles": [
                                    {
                                        "arn": "arn:aws:iam::123456789012:role/role_name",
                                        "account_id": "123456789012",
                                        "account_friendly_name": "prod",
                                        "role_name": "role_name",
                                        "apps": {
                                            "app_details": [
                                                {
                                                    "name": "consoleme",
                                                    "owner": "owner@example.com",
                                                    "owner_url": null,
                                                    "app_url": "https://example.com"
                                                }
                                            ]
                                        }
                                    },
                                    ...
                                ]
                    }
                }
        """
        self.user: str = self.requester["email"]

        include_all_roles = self.get_arguments("all")
        console_only = True
        if include_all_roles == ["true"]:
            console_only = False

        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": self.user,
            "console_only": console_only,
            "message": "Getting all eligible user roles",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }
        log.debug(log_data)
        stats.count("GetRolesMTLSHandler.get", tags={"user": self.user})

        await self.authorization_flow(user=self.user, console_only=console_only)
        eligible_roles_details_array = await get_eligible_role_details(
            sorted(self.eligible_roles)
        )

        res = WebResponse(
            status=Status2.success,
            status_code=200,
            data=eligible_roles_details_array.dict(),
        )
        self.write(res.json(exclude_unset=True))
        await self.finish()



"""Docstring in public module."""
import os
import sys

import ujson as json
from mock import MagicMock, patch
from tornado.testing import AsyncHTTPTestCase

from tests.conftest import MockBaseHandler, MockRedisHandler, create_future

APP_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(APP_ROOT, ".."))

mock_policy_redis = MagicMock(
    return_value=MockRedisHandler(
        return_value={
            "123456789012": (
                '["arn:aws:iam:123456789012:policy/Policy1",'
                '"arn:aws:iam:123456789012:policy/Policy2"]'
            )
        }
    )
)


class TestPolicyResourceEditHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    @patch(
        "consoleme.handlers.v1.policies.ResourceTypeAheadHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    @patch("consoleme.lib.aws.RedisHandler", mock_policy_redis)
    @patch("consoleme.handlers.base.auth")
    @patch("consoleme.handlers.v1.policies.retrieve_json_data_from_redis_or_s3")
    def test_resource_typeahead(
        self, mock_retrieve_json_data_from_redis_or_s3, mock_auth
    ):
        from consoleme.config import config

        mock_auth.validate_certificate.return_value = create_future(True)
        mock_auth.extract_user_from_certificate.return_value = create_future(
            {"name": "user@example.com", "type": "user"}
        )
        mock_auth.get_cert_age_seconds.return_value = create_future(100)
        headers = {
            config.get("auth.user_header_name"): "user@example.com",
            config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        # Invalid resource, no search string
        resource = "fake"
        response = self.fetch(
            f"/api/v1/policies/typeahead?resource={resource}",
            headers=headers,
            method="GET",
        )
        self.assertEqual(response.code, 400)

        # Valid resource, no search string
        resource = "s3"
        response = self.fetch(
            f"/api/v1/policies/typeahead?resource={resource}",
            headers=headers,
            method="GET",
        )
        self.assertEqual(response.code, 400)
        result = create_future({"123456789012": '["abucket1", "abucket2"]'})
        mock_retrieve_json_data_from_redis_or_s3.return_value = result
        account_id = "123456789012"
        resource = "s3"
        search = "a"
        response = self.fetch(
            f"/api/v1/policies/typeahead?resource={resource}&search={search}&account_id={account_id}",
            headers=headers,
            method="GET",
        )
        self.assertEqual(response.code, 200)
        self.assertIsInstance(json.loads(response.body), list)
        self.assertEqual(
            json.loads(response.body),
            [
                {"title": "abucket1", "account_id": "123456789012"},
                {"title": "abucket2", "account_id": "123456789012"},
            ],
        )



import ujson as json
from tornado.testing import AsyncHTTPTestCase


class TestNotFoundHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.config import config

        self.config = config
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_get(self):
        expected = {"status": 404, "title": "Not Found", "message": "Not Found"}
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/route_does_not_exist", method="GET", headers=headers
        )
        self.assertEqual(response.code, 404)
        self.assertDictEqual(json.loads(response.body), expected)

    def test_put(self):
        expected = {"status": 404, "title": "Not Found", "message": "Not Found"}
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/route_does_not_exist", method="PUT", headers=headers, body="{}"
        )
        self.assertEqual(response.code, 404)
        self.assertDictEqual(json.loads(response.body), expected)

    def test_post(self):
        expected = {"status": 404, "title": "Not Found", "message": "Not Found"}
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/route_does_not_exist", method="POST", headers=headers, body="{}"
        )
        self.assertEqual(response.code, 404)
        self.assertDictEqual(json.loads(response.body), expected)

    def test_patch(self):
        expected = {"status": 404, "title": "Not Found", "message": "Not Found"}
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/route_does_not_exist", method="PATCH", headers=headers, body="{}"
        )
        self.assertEqual(response.code, 404)
        self.assertDictEqual(json.loads(response.body), expected)

    def test_delete(self):
        expected = {"status": 404, "title": "Not Found", "message": "Not Found"}
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/route_does_not_exist", method="DELETE", headers=headers
        )
        self.assertEqual(response.code, 404)
        self.assertDictEqual(json.loads(response.body), expected)



"""Group mapping plugin."""
import sys
import time
from typing import List

import sentry_sdk
import simplejson as json
from redis.exceptions import ConnectionError

from consoleme.config import config
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.cloud_credential_authorization_mapping import (
    CredentialAuthorizationMapping,
)
from consoleme.lib.crypto import Crypto
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.redis import RedisHandler

log = config.get_logger("consoleme")
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
credential_authz_mapping = CredentialAuthorizationMapping()

# TODO: Docstrings should provide examples of the data that needs to be returned


class GroupMapping:
    """Group mapping handles mapping groups to eligible roles and accounts."""

    def __init__(self):
        pass

    async def get_eligible_roles(
        self, username: str, groups: list, user_role: str, console_only: bool, **kwargs
    ) -> list:
        """Get eligible roles for user."""
        roles: list = []
        # Legacy cruft, we should rename the parameter here.
        include_cli: bool = not console_only

        roles.extend(
            await credential_authz_mapping.determine_users_authorized_roles(
                username, groups, include_cli
            )
        )

        return list(set(roles))

    @staticmethod
    async def filter_eligible_roles(query: str, obj: object) -> List:
        selected_roles: List = []
        for r in obj.eligible_roles:
            if query.lower() == r.lower():
                # Exact match. Only return the specific role
                return [r]
            if query.lower() in r.lower():
                selected_roles.append(r)
        return list(set(selected_roles))

    async def set_recent_user(self, user):
        pass

    async def set_console_roles_in_cache(
        self,
        user,
        roles,
        expiration=config.get("group_mapping_config.role_cache_expiration", 21600),
    ):
        """Set roles in cache with a nominal expiration"""
        stats.count("set_console_roles_in_cache")
        if not self.red:
            self.red = await RedisHandler().redis()
        expiration = int(time.time()) + expiration
        role_blob = json.dumps({"user": user, "roles": roles, "expiration": expiration})
        crypto = Crypto()
        sig = crypto.sign(role_blob)

        key = config.get(
            "group_mapping_config.role_cache_redis_key", "ROLE_CACHE_{}"
        ).format(user)
        sig_key = config.get(
            "group_mapping_config.role_cache_redis_sig_key", "ROLE_CACHE_SIG_{}"
        ).format(user)

        try:
            self.red.setex(key, expiration, role_blob)
            self.red.setex(sig_key, expiration, sig)
        except ConnectionError:
            log.error("Error connecting to Redis.", exc_info=True)

    async def get_roles_from_cache(self, user):
        """Get roles from cache"""
        stats.count("get_roles_from_cache")

        key = config.get(
            "group_mapping_config.role_cache_redis_key", "ROLE_CACHE_{}"
        ).format(user)
        sig_key = config.get(
            "group_mapping_config.role_cache_redis_sig_key", "ROLE_CACHE_SIG_{}"
        ).format(user)

        if not self.red:
            self.red = await RedisHandler().redis()

        role_r = self.red.get(key)
        if not role_r:
            return []

        role_sig = self.red.get(sig_key)

        if not role_sig:
            stats.count("get_roles_from_cache.no_role_sig")
            log.error("Role data is in redis, but no signature is present.")
            return []

        role_blob = json.loads(role_r)

        if int(time.time()) > role_blob.get("expiration", 0):
            stats.count("get_roles_from_cache.role_cache_expired")
            log.error("Role cache for {} has expired.".format(user))
            return []

        if role_blob.get("user") != user:
            stats.count("get_roles_from_cache.role_cache_user_invalid")
            log.error(
                "Role cache user mismatch. Cache has: {}. User requested is {}".format(
                    role_blob.get("user"), user
                )
            )
            return []
        return role_blob.get("roles")

    async def generate_credential_authorization_mapping(self, authorization_mapping):
        # Override this with company-specific logic
        return authorization_mapping

    async def get_eligible_accounts(self, role_arns):
        """Get eligible accounts for user."""
        stats.count("get_eligible_accounts")
        account_ids = {}

        friendly_names = await get_account_id_to_name_mapping()
        for r in role_arns:
            try:
                account_id = r.split(":")[4]
                account_friendlyname = friendly_names.get(account_id, "")
                if account_friendlyname and isinstance(account_friendlyname, list):
                    account_ids[account_id] = account_friendlyname[0]
                elif account_friendlyname and isinstance(account_friendlyname, str):
                    account_ids[account_id] = account_friendlyname
            except Exception as e:
                log.error(
                    {
                        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
                        "message": "Unable to parse role ARN",
                        "role": r,
                        "error": str(e),
                    }
                )
                sentry_sdk.capture_exception()
        return account_ids

    async def get_account_mappings(self) -> dict:
        """Get a dictionary with all of the account mappings (friendly names -> ID and ID -> names)."""
        return {}

    async def get_secondary_approvers(self, group, return_default=False):
        return config.get("access_requests.default_approver")

    def get_account_names_to_ids(self, force_refresh: bool = False) -> dict:
        """Get account name to id mapping"""
        stats.count("get_account_names_to_ids")
        return {}

    def get_account_ids_to_names(self, force_refresh: bool = False) -> str:
        """Get account id to name mapping"""
        stats.count("get_account_ids_to_names")
        return {}

    async def get_max_cert_age_for_role(self, role_name: str):
        """Retrieve the maximum allowed certificate age allowable to retrieve a particular
        role. 30 will be returned if there is no max age defined.
        """
        return 360

    async def get_all_account_data(self):
        return {}

    async def get_all_accounts(self):
        """Get all account details"""
        return {}

    async def get_all_user_groups(self, user, groups):
        return []

    def is_role_valid(self, entry):
        return True


def init():
    """Initialize group_mapping plugin."""
    return GroupMapping()



import base64

import boto3


def get_aws_secret(secret_name, region):
    session = boto3.session.Session()
    client = session.client(
        service_name="secretsmanager",
        region_name=region,
    )
    get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    if "SecretString" in get_secret_value_response:
        return get_secret_value_response["SecretString"]
    else:
        return base64.b64decode(get_secret_value_response["SecretBinary"])



"""
VPC stack for running ConsoleMe on ECS
"""

import urllib.request

from aws_cdk import aws_ec2 as ec2
from aws_cdk import core as cdk


class VPCStack(cdk.NestedStack):
    """
    VPC stack for running ConsoleMe on ECS
    """

    def __init__(self, scope: cdk.Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        # VPC and security groups

        vpc = ec2.Vpc(self, "Vpc", max_azs=2)

        consoleme_sg = ec2.SecurityGroup(
            self,
            "LBSG",
            vpc=vpc,
            description="ConsoleMe ECS service load balancer security group",
            allow_all_outbound=True,
        )

        # Open ingress to the deploying computer public IP

        my_ip_cidr = (
            urllib.request.urlopen("http://checkip.amazonaws.com")
            .read()
            .decode("utf-8")
            .strip()
            + "/32"
        )

        consoleme_sg.add_ingress_rule(
            peer=ec2.Peer.ipv4(cidr_ip=my_ip_cidr),
            connection=ec2.Port.tcp(port=443),
            description="Allow HTTPS traffic",
        )

        redis_sg = ec2.SecurityGroup(
            self,
            "ECSG",
            vpc=vpc,
            description="ConsoleMe Redis security group",
            allow_all_outbound=True,
        )

        redis_sg.connections.allow_from(
            consoleme_sg,
            port_range=ec2.Port.tcp(port=6379),
            description="Allow ingress from ConsoleMe containers",
        )

        self.vpc = vpc
        self.redis_sg = redis_sg
        self.consoleme_sg = consoleme_sg



from consoleme.config import config
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.plugins import get_plugin_by_name

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()

log = config.get_logger()


class ErrorHandler(BaseAPIV2Handler):
    async def get(self):
        self.write_error(self.get_status())

    async def head(self):
        self.write_error(self.get_status())

    async def put(self):
        self.write_error(self.get_status())

    async def patch(self):
        self.write_error(self.get_status())

    async def post(self):
        self.write_error(self.get_status())

    async def delete(self):
        self.write_error(self.get_status())


class NotFoundHandler(ErrorHandler):
    def initialize(self):
        self.set_status(404)



from typing import Dict

from consoleme.config import config
from consoleme.lib.cloud_credential_authorization_mapping.models import (
    CredentialAuthzMappingGenerator,
    RoleAuthorizations,
    user_or_group,
)
from consoleme.lib.plugins import get_plugin_by_name


class InternalPluginAuthorizationMappingGenerator(CredentialAuthzMappingGenerator):
    async def generate_credential_authorization_mapping(
        self, authorization_mapping: Dict[user_or_group, RoleAuthorizations]
    ) -> Dict[user_or_group, RoleAuthorizations]:
        """This will list accounts that meet the account attribute search criteria."""
        group_mapping = get_plugin_by_name(
            config.get("plugins.group_mapping", "default_group_mapping")
        )()

        # Generate mapping from internal plugin
        authorization_mapping = (
            await group_mapping.generate_credential_authorization_mapping(
                authorization_mapping
            )
        )
        # Return mapping
        return authorization_mapping



import sys

from asgiref.sync import async_to_sync

from consoleme.config import config
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.plugins import get_plugin_by_name

aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()

# ALL ACCOUNTS is a dictionary of account ID to a list of account names (including aliases)
# ex: {"123456": "account_name", ...}
ALL_ACCOUNTS = async_to_sync(get_account_id_to_name_mapping)(status=None)


def format_role_name(arn: str, accounts: dict) -> str:
    """Given a role name, return what to display on the UI. This cleanly formats per-user roles."""

    role_name = arn.split("role/")[1]
    if not accounts:
        # Only fall back to ALL_ACCOUNTS if an accounts dict is not supplied
        accounts = ALL_ACCOUNTS

    if config.get("format_role_name.show_full_arn"):
        return arn
    elif config.get("format_role_name.show_account_name_role_name"):
        account_id = arn.split(":")[4]
        account_name = accounts.get(account_id)
        if not account_name:
            account_name = account_id
        return f"{account_name}/{role_name}"

    if not role_name.startswith("cm_"):
        return role_name

    name = accounts.get(arn.split(":")[4])

    # This should NOT happen, but if it does, log it keep a metric of it:
    if not name:
        log_data = {
            "function": f"{__name__}.{sys._getframe().f_code.co_name}",
            "message": "Can't find account for per-user role",
            "role": role_name,
            "accounts": accounts,
        }
        log.error(log_data)

        stats.count("index.unknown_account_role", tags={"role": role_name})

    return name



from typing import List, Optional

from consoleme.lib.pydantic import BaseModel


class TemplateFile(BaseModel):
    name: Optional[str]
    owner: Optional[str]
    include_accounts: Optional[List[str]]
    exclude_accounts: Optional[List[str]]
    number_of_accounts: Optional[int]
    resource: str
    resource_type: str
    repository_name: str
    template_language: str  # terraform|honeybee
    web_path: str
    file_path: str
    content: Optional[str]


class TemplatedFileModelArray(BaseModel):
    templated_resources: List[TemplateFile]



import sys

import sentry_sdk
from pydantic import ValidationError

from consoleme.config import config
from consoleme.exceptions.exceptions import InvalidRequestParameter
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.change_request import generate_change_model_array
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import ChangeGeneratorModelArray

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


class GenerateChangesHandler(BaseAPIV2Handler):
    """Handler for /api/v2/generate_changes

    Generates a ChangeModelArray from ChangeGeneratorModelArray
    """

    allowed_methods = ["POST"]

    async def post(self):
        """
        POST /api/v2/generate_changes

        Generates a ChangeModelArray JSON from ChangeGeneratorModelArray JSON.

        Request example:

        {"changes": [
            {
                "principal": {
                    "principal_arn": "arn:aws:iam::123456789012:role/aRole",
                    "principal_type": "AwsResource"
                },
                "generator_type": "s3",
                "resource_arn": ["arn:aws:s3:::123456789012-bucket"],
                "bucket_prefix": "/*",
                "effect": "Allow",
                "action_groups": [
                    "get",
                    "list"
                ]
            }
        ]}

        Response example:

        { "changes" : [
                {
                    "principal": {
                        "principal_arn": "arn:aws:iam::123456789012:role/aRole",
                        "principal_type": "AwsResource"
                    },
                    "change_type": "inline_policy",
                    "resource_arn": [
                        "arn:aws:s3:::123456789012-bucket"
                    ],
                    "resource": null,
                    "condition": null,
                    "policy_name": "cm_user_1592499820_gmli",
                    "new": true,
                    "policy": {
                        "version": "2012-10-17",
                        "statements": null,
                        "policy_document": "{\"Version\":\"2012-10-17\",\"Statement\":[[{\"Action\"...",
                        "policy_sha256": "cb300def8dd1deaf4db2bfeef4bc6fc740be18e8ccae74c399affe781f82ba6e"
                    },
                    "old_policy": null
                }
            ]
        }
        """

        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "ip": self.ip,
            "request_id": self.request_uuid,
        }

        try:
            # Validate the model
            changes = ChangeGeneratorModelArray.parse_raw(self.request.body)

            # Override user attribute for each change
            for change in changes.changes:
                change.user = self.user

            # Loop through the raw json object to retrieve attributes that would be parsed out in the
            # ChangeGeneratorModelArray, such as bucket_prefix for S3ChangeGeneratorModel
            change_model_array = await generate_change_model_array(changes)
        except (InvalidRequestParameter, ValidationError) as e:
            log_data["message"] = "Validation Exception"
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.validation_exception", tags={"user": self.user}
            )
            self.write_error(400, message="Error validating input: " + str(e))
            return
        except Exception as e:
            log_data["message"] = "Unknown Exception occurred while generating changes"
            log.error(log_data, exc_info=True)
            stats.count(f"{log_data['function']}.exception", tags={"user": self.user})
            sentry_sdk.capture_exception(tags={"user": self.user})
            self.write_error(500, message="Error generating changes: " + str(e))
            return

        log_data["message"] = "Successfully generated changes requested"
        log.info(log_data)
        stats.count(f"{log_data['function']}.success", tags={"user": self.user})
        self.write(change_model_array.json())
